import decimal
import pickle
import textwrap

from . import types

# Hashable types. Checking ``issubclass(T, types.Hashable)`` does not
# work because named tuples pass that test and beancount uses many named
# tuples that have dictionary members making them effectively not
# hashable.
FUNDAMENTAL = frozenset({
    bool,
    bytes,
    complex,
    decimal.Decimal,
    float,
    int,
    str,
    # These are hashable only if the contained objects are hashable.
    frozenset,
    tuple,
})

# Function reducing non-hashable types to something hashable.
REDUCERS = {}


def register(datatype, func):
    """Register reduce function for non-hashable type.

    The reduce function maps an non-hashable object into an hashable
    representation. This representation does not need to capture all the
    object facets, but it should retrurn someting unique enough to avoid
    too many hashing collisions.
    """
    REDUCERS[datatype] = func


def make(columns):
    """Build an hashable tuple subclass."""

    # When all columns are hashable, pass the input tuple through as is.
    if all(column.datatype in FUNDAMENTAL for column in columns):
        return lambda x: x

    datatypes = ', '.join(types.name(column.datatype) for column in columns)

    # Code generation inspired by standard library ``dataclasses.py``.
    parts = []
    locals = {}
    for i, column in enumerate(columns):
        if column.datatype in FUNDAMENTAL:
            parts.append(f'self[{i}]')
        elif column.datatype is dict:
            parts.append(f'*self[{i}].keys(), *self[{i}].values()')
        elif column.datatype is set:
            parts.append(f'*self[{i}]')
        else:
            func = REDUCERS.get(column.datatype, pickle.dumps)
            fname = f'func{i}'
            locals[fname] = func
            parts.append(f'{fname}(self[{i}])')

    objs = ', '.join(parts)
    names = ', '.join(locals.keys())
    code = textwrap.dedent(f'''
        def create({names}):
            def __hash__(self):
                return hash(({objs}))
            return __hash__
    ''')
    clsname = f'Hashable[{datatypes}]'

    ns = {}
    exec(code, globals(), ns)
    func = ns['create'](**locals)
    func.__qualname__ = f'{clsname}.{func.__name__}'

    members = dict(tuple.__dict__)
    members['__hash__'] = func

    return type(clsname, (tuple,), members)



import copy
import datetime
import sys
import types as _types
import typing

from decimal import Decimal
from functools import lru_cache as cache
from urllib.parse import urlparse

from beancount import loader
from beancount import parser
from beancount.core import amount
from beancount.core import convert
from beancount.core import data
from beancount.core import getters
from beancount.core import inventory
from beancount.core import position
from beancount.core import prices
from beancount.core.compare import hash_entry
from beancount.core.getters import get_account_open_close, get_commodity_directives
from beancount.ops import summarize

from beanquery import tables
from beanquery import types
from beanquery import query_compile
from beanquery import query_render
from beanquery import hashable
from beanquery.query_env import ColumnsRegistry


if sys.version_info >= (3, 10):
    _UNIONTYPES = {typing.Union, _types.UnionType}
else:
    _UNIONTYPES = {typing.Union}


hashable.register(data.Transaction, lambda t: (t.date, t.flag, t.narration))


_TABLES = []


def attach(context, dsn, *, entries=None, errors=None, options=None):
    filename = urlparse(dsn).path
    if filename:
        entries, errors, options = loader.load_file(filename)
    for table in _TABLES:
        context.tables[table.name] = table(entries, options)
    context.options.update(options)
    context.errors.extend(errors)
    # Set the default table. This eventually will have to be removed.
    context.tables[None] = context.tables['postings']


class Metadata(dict):
    pass


class MetadataRenderer(query_render.ObjectRenderer):
    dtype = Metadata

    def format(self, value):
        return str({k: v for k, v in value.items() if k not in {'filename', 'lineno'} and not k.startswith('__')})


class GetAttrColumn(query_compile.EvalColumn):
    def __init__(self, name, dtype):
        super().__init__(dtype)
        self.name = name

    def __call__(self, context):
        return getattr(context, self.name)


def _simplify_typing_annotation(dtype):
    if typing.get_origin(dtype) in _UNIONTYPES:
        args = typing.get_args(dtype)
        if len(args) == 2:
            if args[0] is type(None):
                return args[1], True
            if args[1] is type(None):
                return args[0], True
        raise NotImplementedError
    return typing.get_origin(dtype) or dtype, False


def _typed_namedtuple_to_columns(cls, renames=None):
    columns = {}
    for name, dtype in typing.get_type_hints(cls).items():
        dtype, nullable = _simplify_typing_annotation(dtype)
        if name == 'meta' and dtype is dict:
            dtype = Metadata
        if dtype is frozenset:
            dtype = set
        colname = renames.get(name, name) if renames is not None else name
        columns[colname] = GetAttrColumn(name, dtype)
    return columns


class Position(types.Structure):
    name = 'position'
    columns = _typed_namedtuple_to_columns(position.Position)


class Cost(types.Structure):
    name = 'cost'
    columns = _typed_namedtuple_to_columns(data.Cost)


class Amount(types.Structure):
    name = 'amount'
    columns = _typed_namedtuple_to_columns(data.Amount)


class Transaction(types.Structure):
    name = 'transaction'
    columns = ColumnsRegistry(_typed_namedtuple_to_columns(data.Transaction))
    del columns['postings']

    @columns.register(typing.Set[str])
    def accounts(row):
        return {p.account for p in row.postings}


types.ALIASES[position.Position] = Position
types.ALIASES[data.Cost] = Cost
types.ALIASES[data.Amount] = Amount
types.ALIASES[data.Transaction] = Transaction


class Open(types.Structure):
    name = 'open'
    columns = _typed_namedtuple_to_columns(data.Open)


class Close(types.Structure):
    name = 'close'
    columns = _typed_namedtuple_to_columns(data.Close)


class Table(tables.Table):
    datatype = None

    def __init__(self, entries, options):
        self.entries = entries
        self.options = options

    def __init_subclass__(cls):
        _TABLES.append(cls)

    def __iter__(self):
        datatype = self.datatype
        for entry in self.entries:
            if isinstance(entry, datatype):
                yield entry

    @property
    def wildcard_columns(self):
        return tuple(col for col in self.columns.keys() if col != 'meta')


class TransactionsTable(Table):
    name = 'transactions'
    datatype = data.Transaction
    # There is not a way to inherit the __init_subclass__() and
    # __iter__() methods while inheriting the columns attribute
    # definition from the Transaction class.
    columns = Transaction.columns


class PricesTable(Table):
    name = 'prices'
    datatype = data.Price
    columns = _typed_namedtuple_to_columns(datatype)

    def __init__(self, entries, options):
        super().__init__(entries, options)
        self.price_map = prices.build_price_map(entries)


class BalancesTable(Table):
    name = 'balances'
    datatype = data.Balance
    columns = _typed_namedtuple_to_columns(datatype, {'diff_amount': 'discrepancy'})


class NotesTable(Table):
    name = 'notes'
    datatype = data.Note
    columns = _typed_namedtuple_to_columns(datatype)


class EventsTable(Table):
    name = 'events'
    datatype = data.Event
    columns = _typed_namedtuple_to_columns(datatype)


class DocumentsTable(Table):
    name = 'documents'
    datatype = data.Document
    columns = _typed_namedtuple_to_columns(datatype)


class GetItemColumn(query_compile.EvalColumn):
    def __init__(self, key, dtype):
        super().__init__(dtype)
        self.key = key

    def __call__(self, row):
        return row[self.key]


class AccountsTable(tables.Table):
    name = 'accounts'
    columns = {
        'account': GetItemColumn(0, str),
        'open': GetItemColumn(1, Open),
        'close': GetItemColumn(2, Close),
    }

    def __init__(self, entries, options):
        self.accounts = get_account_open_close(entries)
        self.types = parser.options.get_account_types(options)

    def __iter__(self):
        return ((name, value[0], value[1]) for name, value in self.accounts.items())

_TABLES.append(AccountsTable)


class CommoditiesTable(tables.Table):
    name = 'commodities'
    columns = _typed_namedtuple_to_columns(data.Commodity, {'currency': 'name'})

    def __init__(self, entries, options):
        self.commodities = get_commodity_directives(entries)

    def __iter__(self):
        return iter(self.commodities.values())

_TABLES.append(CommoditiesTable)


class _BeancountTable(tables.Table):
    def __init__(self, entries, options, open=None, close=None, clear=None):
        super().__init__()
        self.entries = entries
        self.options = options
        self.open = open
        self.close = close
        self.clear = clear

    def evolve(self, **kwargs):
        table = copy.copy(self)
        for name, value in kwargs.items():
            setattr(table, name, value)
        return table

    def prepare(self):
        """Filter the entries applying the FROM clause qualifiers OPEN, CLOSE, CLEAR."""
        entries = self.entries
        options = self.options

        # Process the OPEN clause.
        if self.open is not None:
            entries, index = summarize.open_opt(entries, self.open, options)

        # Process the CLOSE clause.
        if self.close is not None:
            if isinstance(self.close, datetime.date):
                entries, index = summarize.close_opt(entries, self.close, options)
            elif self.close is True:
                entries, index = summarize.close_opt(entries, None, options)

        # Process the CLEAR clause.
        if self.clear is not None:
            entries, index = summarize.clear_opt(entries, None, options)

        return entries


class EntriesTable(_BeancountTable):
    name = 'entries'
    columns = ColumnsRegistry()

    def __iter__(self):
        entries = self.prepare()
        yield from iter(entries)

    @columns.register(str)
    def id(entry):
        """Unique id of a directive."""
        return hash_entry(entry)

    @columns.register(str)
    def type(entry):
        """The data type of the directive."""
        return type(entry).__name__.lower()

    @columns.register(str)
    def filename(entry):
        """The filename where the directive was parsed from or created."""
        return entry.meta["filename"]

    @columns.register(int)
    def lineno(entry):
        """The line number from the file the directive was parsed from."""
        return entry.meta["lineno"]

    @columns.register(datetime.date)
    def date(entry):
        """The date of the directive."""
        return entry.date

    @columns.register(int)
    def year(entry):
        """The year of the date year of the directive."""
        return entry.date.year

    @columns.register(int)
    def month(entry):
        """The year of the date month of the directive."""
        return entry.date.month

    @columns.register(int)
    def day(entry):
        """The year of the date day of the directive."""
        return entry.date.day

    @columns.register(str)
    def flag(entry):
        """The flag the transaction."""
        if not isinstance(entry, data.Transaction):
            return None
        return entry.flag

    @columns.register(str)
    def payee(entry):
        """The payee of the transaction."""
        if not isinstance(entry, data.Transaction):
            return None
        return entry.payee

    @columns.register(str)
    def narration(entry):
        """The narration of the transaction."""
        if not isinstance(entry, data.Transaction):
            return None
        return entry.narration

    @columns.register(str)
    def description(entry):
        """A combination of the payee + narration of the transaction, if present."""
        if not isinstance(entry, data.Transaction):
            return None
        return ' | '.join(filter(None, [entry.payee, entry.narration]))

    @columns.register(set)
    def tags(entry):
        """The set of tags of the transaction."""
        return getattr(entry, 'tags', None)

    @columns.register(set)
    def links(entry):
        """The set of links of the transaction."""
        return getattr(entry, 'links', None)

    @columns.register(dict)
    def meta(entry):
        return entry.meta

    @columns.register(typing.Set[str])
    def accounts(entry):
        return getters.get_entry_accounts(entry)

_TABLES.append(EntriesTable)


class _PostingsTableRow:
    """A dumb container for information used by a row expression."""

    def __init__(self):
        self.rowid = 0
        self.balance = inventory.Inventory()

        # The current transaction of the posting being evaluated.
        self.entry = None

        # The current posting being evaluated.
        self.posting = None

    def __hash__(self):
        # The context hash is used in caching column accessor functions.
        # Instead than hashing the row context content, use the rowid as
        # hash.
        return self.rowid


class PostingsTable(_BeancountTable):
    name = 'postings'
    columns = ColumnsRegistry()
    wildcard_columns = 'date flag payee narration position'.split()

    def __iter__(self):
        entries = self.prepare()
        context = _PostingsTableRow()
        for entry in entries:
            if isinstance(entry, data.Transaction):
                context.entry = entry
                for posting in entry.postings:
                    context.rowid += 1
                    context.posting = posting
                    yield context

    @columns.register(str)
    def type(context):
        return 'transaction'

    @columns.register(str)
    def id(context):
        """Unique id of a directive."""
        return hash_entry(context.entry)

    @columns.register(datetime.date)
    def date(context):
        """The date of the directive."""
        return context.entry.date

    @columns.register(int)
    def year(context):
        """The year of the date year of the directive."""
        return context.entry.date.year

    @columns.register(int)
    def month(context):
        """The year of the date month of the directive."""
        return context.entry.date.month

    @columns.register(int)
    def day(context):
        """The year of the date day of the directive."""
        return context.entry.date.day

    @columns.register(str)
    def filename(context):
        """The ledger where the posting is defined."""
        meta = context.posting.meta
        # Postings for pad transactions have their meta fields set to
        # None. See https://github.com/beancount/beancount/issues/767
        if meta is None:
            return None
        return meta["filename"]

    @columns.register(int)
    def lineno(context):
        """The line number in the ledger file where the posting is defined."""
        meta = context.posting.meta
        # Postings for pad transactions have their meta fields set to
        # None. See https://github.com/beancount/beancount/issues/767
        if meta is None:
            return None
        return meta["lineno"]

    @columns.register(str)
    def location(context):
        """The filename:lineno location where the posting is defined."""
        meta = context.posting.meta
        # Postings for pad transactions have their meta fields set to
        # None. See https://github.com/beancount/beancount/issues/767
        if meta is None:
            return None
        return '{:s}:{:d}:'.format(meta['filename'], meta['lineno'])

    @columns.register(str)
    def flag(context):
        """The flag of the parent transaction for this posting."""
        return context.entry.flag

    @columns.register(str)
    def payee(context):
        """The payee of the parent transaction for this posting."""
        return context.entry.payee

    @columns.register(str)
    def narration(context):
        """The narration of the parent transaction for this posting."""
        return context.entry.narration

    @columns.register(str)
    def description(context):
        """A combination of the payee + narration for the transaction of this posting."""
        return ' | '.join(filter(None, [context.entry.payee, context.entry.narration]))

    @columns.register(set)
    def tags(context):
        """The set of tags of the parent transaction for this posting."""
        return context.entry.tags

    @columns.register(set)
    def links(context):
        """The set of links of the parent transaction for this posting."""
        return context.entry.links

    @columns.register(str)
    def posting_flag(context):
        """The flag of the posting itself."""
        return context.posting.flag

    @columns.register(str)
    def account(context):
        """The account of the posting."""
        return context.posting.account

    @columns.register(set)
    def other_accounts(context):
        """The list of other accounts in the transaction, excluding that of this posting."""
        return sorted({posting.account for posting in context.entry.postings if posting is not context.posting})

    @columns.register(Decimal)
    def number(context):
        """The number of units of the posting."""
        return context.posting.units.number

    @columns.register(str)
    def currency(context):
        """The currency of the posting."""
        return context.posting.units.currency

    @columns.register(Decimal)
    def cost_number(context):
        """The number of cost units of the posting."""
        cost = context.posting.cost
        return cost.number if cost else None

    @columns.register(str)
    def cost_currency(context):
        """The cost currency of the posting."""
        cost = context.posting.cost
        return cost.currency if cost else None

    @columns.register(datetime.date)
    def cost_date(context):
        """The cost currency of the posting."""
        cost = context.posting.cost
        return cost.date if cost else None

    @columns.register(str)
    def cost_label(context):
        """The cost currency of the posting."""
        cost = context.posting.cost
        return cost.label if cost else ''

    @columns.register(position.Position)
    def position(context):
        """The position for the posting. These can be summed into inventories."""
        posting = context.posting
        return position.Position(posting.units, posting.cost)

    @columns.register(amount.Amount)
    def price(context):
        """The price attached to the posting."""
        return context.posting.price

    @columns.register(amount.Amount)
    def weight(context):
        """The computed weight used for this posting."""
        return convert.get_weight(context.posting)

    @columns.register(inventory.Inventory)
    @cache(maxsize=1)  # noqa: B019
    def balance(context):
        """The balance for the posting. These can be summed into inventories."""
        # Caching protects against multiple balance updates per row when
        # the columns appears more than once in the execurted query. The
        # rowid in the row context guarantees that otherwise identical
        # rows do not hit the cache and thus that the balance is correctly
        # updated.
        context.balance.add_position(context.posting)
        return copy.copy(context.balance)

    @columns.register(dict)
    def meta(context):
        return context.posting.meta

    @columns.register(data.Transaction)
    def entry(context):
        return context.entry

    @columns.register(typing.Set[str])
    def accounts(context):
        return {p.account for p in context.entry.postings}

_TABLES.append(PostingsTable)



import datetime
import decimal
import itertools
import typing


# Only Python >= 3.10 exposes NoneType in the types module.
NoneType = type(None)


class AnyType:
    """As written on the tin."""
    __slots__ = ()
    __name__ = 'any'

    def __eq__(self, other):
        """Compares equal to any other type."""
        return isinstance(other, type)


# Used in BQL functions signatures for arguments that can have any type.
Any = AnyType()


# Used for COUNT(*)
Asterisk = typing.NewType('*', object)  # noqa: PLC0132
Asterisk.__mro__ = Asterisk,


# Keep track of the defined structured types to allow introspection.
TYPES = {}


class Structure:
    """Base class for structured data types."""
    name = None
    columns = {}

    def __init_subclass__(cls):
        if cls.name:
            TYPES[cls.name] = cls


def _bases(t):
    if t is NoneType:
        return (object,)
    bases = t.__mro__
    if len(bases) > 1 and bases[-1] is object:
        # All types that are not ``object`` have more than one class
        # in their ``__mro__``. BQL uses ``object`` for untypes
        # values. Do not return ``object`` as base for strict types,
        # to avoid functions taking untyped onjects to accept all
        # values.
        return bases[:-1]
    return bases


def function_lookup(functions, name, operands):
    """Lookup a BQL function implementation.

    Args:
      functions: The functions registry to interrogate.
      name: The function name.
      operands: Function operands.

    Returns:
      A EvalNode (or subclass) instance or None if the function was not found.
    """
    for signature in itertools.product(*(_bases(operand.dtype) for operand in operands)):
        for func in functions[name]:
            if func.__intypes__ == list(signature):
                return func
    return None


# Map types to their BQL name. Used to find the name of the type cast funtion.
MAP = {
    bool: 'bool',
    datetime.date: 'date',
    decimal.Decimal: 'decimal',
    int: 'int',
    str: 'str',
}


# Map between Python types and BQL structured types. Functions and
# columns definitions can use Python types. The corresponding BQL
# structured type is looked up when compiling the subscrip operator.
ALIASES = {}


def name(datatype):
    if datatype is NoneType:
        return 'NULL'
    if isinstance(datatype, typing._GenericAlias):
        return str(datatype).rsplit('.', 1)[-1].lower()
    return getattr(datatype, 'name', datatype.__name__.lower())


_STRING_TO_DATATYPE = {
    'bool': bool,
    'date': datetime.date,
    'decimal': decimal.Decimal,
    'int': int,
    'object': object,
    'str': str,
    'text': str,
    'varchar': str,
}


def parse(name):
    """Parse the string representation of a type into a type object.

    This does not (yet) work for all supprted data types.
    """
    return _STRING_TO_DATATYPE.get(name, None)



import importlib

from urllib.parse import urlparse

from . import parser
from . import compiler
from . import tables

from .compiler import CompilationError
from .cursor import Cursor, Column
from .errors import Warning, Error, InterfaceError, DatabaseError, DataError, OperationalError
from .errors import IntegrityError, InternalError, ProgrammingError, NotSupportedError
from .parser import ParseError


__version__ = '0.2.0'


# DB-API compliance
apilevel = '2.0'
threadsafety = 2
paramstyle = 'pyformat'


def connect(dsn, **kwargs):
    return Connection(dsn, **kwargs)


class Connection:
    def __init__(self, dsn='', **kwargs):
        self.tables = {None: tables.NullTable()}

        # The ``None`` table is the default table. The ``''`` table is the
        # table that is explicitly selected with ``FROM #``. Having the
        # default table and the ``''`` table allows to select the empty table
        # when the ``beancount`` data source is initialized and it sets the
        # default table to the ``postings`` table.
        self.tables[''] = self.tables[None]

        # These are used only by the ``beancount`` data source.
        self.options = {}
        self.errors = []

        if dsn:
            self.attach(dsn, **kwargs)

    def attach(self, dsn, **kwargs):
        scheme = urlparse(dsn).scheme
        source = importlib.import_module(f'beanquery.sources.{scheme}')
        source.attach(self, dsn, **kwargs)

    def close(self):
        # Required by the DB-API.
        pass

    def parse(self, query):
        return parser.parse(query)

    def compile(self, query):
        return compiler.compile(self, query)

    def execute(self, query, params=None):
        return self.cursor().execute(query, params)

    def cursor(self):
        return Cursor(self)


__all__ = [
    'Column',
    'CompilationError',
    'Connection',
    'Cursor',
    'DataError',
    'DatabaseError',
    'Error',
    'IntegrityError',
    'InterfaceError',
    'InternalError',
    'NotSupportedError',
    'OperationalError',
    'ParseError',
    'ProgrammingError',
    'Warning',
    'apilevel',
    'connet',
    'paramstyle',
    'threadsafety',
]



"""A library to run queries. This glues together all the parts of the query engine.
"""
__copyright__ = "Copyright (C) 2015-2017  Martin Blais"
__license__ = "GNU GPLv2"

import beanquery
import beanquery.numberify


def run_query(entries, options, query, *args, numberify=False):
    """Compile and execute a query, return the result types and rows.

    Args:
      entries: A list of entries, as produced by the loader.
      options: A dict of options, as produced by the loader.
      query: A string, a single BQL query, optionally containing some new-style
        (e.g., {}) formatting specifications.
      args: A tuple of arguments to be formatted in the query. This is
        just provided as a convenience.
      numberify: If true, numberify the results before returning them.
    Returns:
      A pair of result types and result rows.
    Raises:
      ParseError: If the statement cannot be parsed.
      CompilationError: If the statement cannot be compiled.
    """

    # Execute the query.
    ctx = beanquery.connect('beancount:', entries=entries, errors=[], options=options)
    curs = ctx.execute(query.format(*args))
    rrows = curs.fetchall()
    rtypes = curs.description

    # Numberify the results, if requested.
    if numberify:
        dformat = options['dcontext'].build()
        rtypes, rrows = beanquery.numberify.numberify_results(rtypes, rrows, dformat)

    return rtypes, rrows



import dataclasses
import unittest

from beanquery import hashable
from beanquery.cursor import Column


class TestHashable(unittest.TestCase):

    def test_fundamental(self):
        columns = (Column('b', bool), Column('i', int), Column('s', str))
        wrap = hashable.make(columns)
        obj = (True, 42, 'universe')
        self.assertIs(wrap(obj), obj)
        hash(obj)

    def test_dict(self):
        columns = (Column('b', bool), Column('d', dict))
        wrap = hashable.make(columns)
        obja = (True, {'answer': 42})
        a = hash(wrap(obja))
        objb = (True, {'answer': 42})
        b = hash(wrap(objb))
        self.assertIsNot(obja, objb)
        self.assertEqual(a, b)
        objc = (False, {'answer': 42})
        c = hash(wrap(objc))
        self.assertNotEqual(a, c)
        objd = (True, {'answer': 43})
        d = hash(wrap(objd))
        self.assertNotEqual(a, d)

    def test_registered(self):

        @dataclasses.dataclass
        class Foo:
            xid: int
            meta: dict

        hashable.register(Foo, lambda obj: obj.xid)

        columns = (Column('b', bool), Column('foo', Foo))
        wrap = hashable.make(columns)
        obja = (True, Foo(1, {'test': 1}))
        a = hash(wrap(obja))
        objb = (True, Foo(1, {'test': 2}))
        b = hash(wrap(objb))
        self.assertEqual(a, b)
        objc = (True, Foo(2, {'test': 2}))
        c = hash(wrap(objc))
        self.assertNotEqual(a, c)



__copyright__ = "Copyright (C) 2014-2016  Martin Blais"
__license__ = "GNU GPLv2"

import datetime
import enum
import io
import textwrap
import unittest

from decimal import Decimal

from beancount.core import display_context
from beancount.core.amount import Amount, A
from beancount.core.inventory import Inventory, from_string as I
from beancount.core.number import D
from beancount.core.position import Cost, Position, from_string as P

from beanquery import query_render
from beanquery.cursor import Column


class TestColumnRenderer(unittest.TestCase):

    def test_column_renderer(self):
        dcontext = display_context.DisplayContext()
        ctx = query_render.RenderContext(dcontext)
        renderer = query_render.ColumnRenderer(ctx)
        with self.assertRaises(RuntimeError):
            w = renderer.width
        w = renderer.prepare()
        self.assertEqual(w, 0)
        self.assertEqual(renderer.width, 0)
        with self.assertRaises(NotImplementedError):
            renderer.format(None)


class RendererTestBase(unittest.TestCase):

    def setUp(self):
        self.dcontext = display_context.DisplayContext()

    def render(self, dtype, values, **kwargs):
        out = io.StringIO()
        rows = [(x, ) for x in values]
        query_render.render_text([Column('', dtype)], rows, self.dcontext, out, **kwargs)
        return out.getvalue().splitlines()[2:]


class Foo(enum.Enum):
    SHORT = 1
    LONG = 2


class TestRenderer(RendererTestBase):

    def test_object(self):
        self.assertEqual(self.render(object, ["foo", 1, D('1.23'), datetime.date(1970, 1, 1)]), [
            'foo       ',
            '1         ',
            '1.23      ',
            '1970-01-01',
        ])

    def test_bool(self):
        self.assertEqual(self.render(bool, [True, True]), [
            'TRUE',
            'TRUE',
        ])
        self.assertEqual(self.render(bool, [False, True]), [
            'FALSE',
            'TRUE ',
        ])
        self.assertEqual(self.render(bool, [False, False]), [
            'FALSE',
            'FALSE',
        ])

    def test_str(self):
        self.assertEqual(self.render(str, ['a', 'bb', 'ccc', '']), [
            'a  ',
            'bb ',
            'ccc',
            '   ',
        ])

    def test_str_null(self):
        self.assertEqual(self.render(str, ['', None]), [
            ' ',
            ' ',
        ])
        self.assertEqual(self.render(str, ['', None], nullvalue='NULL'), [
            '    ',
            'NULL',
        ])

    def test_set_str(self):
        self.assertEqual(self.render(set, [{}, {'aaaa'}, {'bb', 'ccc'}]), [
            '       ',
            'aaaa   ',
            'bb  ccc',
        ])

    def test_date(self):
        self.assertEqual(self.render(datetime.date, [datetime.date(2014, 10, 3)]), [
            '2014-10-03'
        ])

    def test_int(self):
        self.assertEqual(self.render(int, [1, 22, 333]), [
            '  1',
            ' 22',
            '333',
        ])
        self.assertEqual(self.render(int, [1, -22, 333]), [
            '  1',
            '-22',
            '333',
        ])
        self.assertEqual(self.render(int, [1, 22, -333]), [
            '   1',
            '  22',
            '-333',
        ])

    def test_decimal_integral(self):
        self.assertEqual(self.render(Decimal, [D('1'), D('12'), D('123'), D('1e4')]), [
            '   1',
            '  12',
            ' 123',
            '1E+4',
        ])
        self.assertEqual(self.render(Decimal, [D('1'), D('-12'), D('123')]), [
            '  1',
            '-12',
            '123',
        ])
        self.assertEqual(self.render(Decimal, [D('1'), D('12'), D('-123')]), [
            '   1',
            '  12',
            '-123',
        ])
        self.assertEqual(self.render(Decimal, [D('1'), D('12'), D('-1e3')]), [
            '    1',
            '   12',
            '-1E+3',
        ])

    def test_decimal_fractional(self):
        self.assertEqual(self.render(Decimal, [D('0.1'), D('1.2'), D('1.23'), D('1.234')]), [
            '0.1  ',
            '1.2  ',
            '1.23 ',
            '1.234',
        ])
        self.assertEqual(self.render(Decimal, [D('12'), D('1.2'), D('1.23'), D('12.345')]), [
            '12    ',
            ' 1.2  ',
            ' 1.23 ',
            '12.345',
        ])
        self.assertEqual(self.render(Decimal, [D('12'), D('1.2'), D('1.23'), D('-12.345')]), [
            ' 12    ',
            '  1.2  ',
            '  1.23 ',
            '-12.345',
        ])
        self.assertEqual(self.render(Decimal, [D('12'), D('1.2'), D('1.23'), D('-1.2e3')]), [
            '     12   ',
            '      1.2 ',
            '      1.23',
            '-1.2E+3   ',
        ])

    def test_enum(self):
        self.assertEqual(self.render(Foo, [Foo.SHORT, Foo.LONG]), [
            'SHORT',
            'LONG ',
        ])


class TestAmountRenderer(RendererTestBase):

    def render(self, values, **kwargs):
        return super().render(Amount, values, **kwargs)

    def test_amount(self):
        self.assertEqual(
            self.render([A('100.00 USD')]), ['100.00 USD'])

    def test_quantization_one(self):
        self.dcontext.update(Decimal('1.0000'), 'ETH')
        self.assertEqual(self.dcontext.quantize(Decimal('1.0'), 'ETH'), Decimal('1.0000'))
        self.assertEqual(self.render([A('1 ETH')]), ['1.0000 ETH'])
        self.assertEqual(self.render([A('0.00001 ETH')]), ['0.0000 ETH'])

    def test_quantization_many(self):
        self.dcontext.update(Decimal('1.0000'), 'ETH')
        self.dcontext.update(Decimal('1.00'), 'USD')
        self.dcontext.update(Decimal('1'), 'XYZ')
        self.assertEqual(self.render([A('1.0 ETH')]), ['1.0000 ETH'])
        self.assertEqual(self.render([A('1.0 USD')]), ['1.00 USD'])
        self.assertEqual(self.render([A('1.0 XYZ')]), ['1 XYZ'])

    def test_number_padding(self):
        # FIXME: The leading space seems like a bug in
        # DisplayContext. Either it should always be there or it
        # should be there to support minus signs not encountered in
        # training or it shoudl not be there at all.
        self.assertEqual(self.render([A('1 XY'), A('12 XY'), A('123 XY'), A('-1 XY')]), [
            '   1 XY',
            '  12 XY',
            ' 123 XY',
            '  -1 XY',
        ])
        self.assertEqual(self.render([A('1 XY'), A('12 XY'), A('-12 XY')]), [
            '  1 XY',
            ' 12 XY',
            '-12 XY',
        ])

    def test_decimal_alignment(self):
        self.assertEqual(self.render([A('1.0 AA'), A('1.00 BB'), A('1.000 CC')]), [
            '1.0   AA',
            '1.00  BB',
            '1.000 CC',
        ])

    def test_currency_padding(self):
        self.assertEqual(self.render([A('1.00 XY'), A('1.00 XYZ'), A('1.00 XYZK')]), [
            '1.00 XY  ',
            '1.00 XYZ ',
            '1.00 XYZK',
        ])

    def test_many(self):
        self.assertEqual(self.render([A('0.0001 USD'), A('20.002 HOOL'), A('33 CA'), A('1098.20 AAPL')]), [
            '   0.0001 USD ',
            '  20.002  HOOL',
            '  33      CA  ',
            '1098.20   AAPL',
        ])


class TestPositionRenderer(RendererTestBase):

    def render(self, values, **kwargs):
        return super().render(Position, values, **kwargs)

    def setUp(self):
        super().setUp()
        # Prime the display context for some known commodities.
        self.dcontext = display_context.DisplayContext()
        self.dcontext.update(D('1.00'), 'USD')
        self.dcontext.update(D('1.00'), 'CAD')
        self.dcontext.update(D('1.000'), 'HOOL')
        self.dcontext.update(D('1'), 'CA')
        self.dcontext.update(D('1.00'), 'AAPL')

    def test_simple_poitions(self):
        self.assertEqual(self.render([P('3.0 USD'), P('3.0 CAD'), P('3.0 HOOL'), P('3.0 CA'), P('3.0 AAPL'), P('3.0 XY')]), [
            '3.00  USD ',
            '3.00  CAD ',
            '3.000 HOOL',
            '3     CA  ',
            '3.00  AAPL',
            '3.0   XY  ',
        ])

    def test_positions_with_price(self):
        self.assertEqual(self.render([P('5 HOOL {500.230000 USD}'), P('123.0 CA {1 HOOL}')]), [
            '  5.000 HOOL {500.23  USD }',
            '123     CA   {  1.000 HOOL}',
        ])


class TestInventoryRenderer(RendererTestBase):

    def render(self, values, **kwargs):
        return super().render(Inventory, values, **kwargs)

    def setUp(self):
        super().setUp()
        # Prime the display context for some known commodities.
        self.dcontext = display_context.DisplayContext()
        self.dcontext.update(D('1.00'), 'USD')
        self.dcontext.update(D('1.00'), 'CAD')
        self.dcontext.update(D('1.000'), 'HOOL')
        self.dcontext.update(D('1'), 'CA')
        self.dcontext.update(D('1.00'), 'AAPL')

    def test_position_sortkey(self):
        inventory = I('1 AAAAA, 5 SHARE {100 USD}, 5 SHARE {200 USD}, 5 TESTS {666 USD}')
        self.assertEqual(
            sorted(inventory, key=query_render.InventoryRenderer.positionsortkey), [
                P('1 AAAAA'),
                P('5 SHARE {200 USD}'),
                P('5 SHARE {100 USD}'),
                P('5 TESTS {666 USD}'),
            ])

    def test_inventory(self):
        self.assertEqual(self.render([I('100 USD')], expand=True),[
            '100.00 USD'
        ])
        self.assertEqual(self.render([I('5 HOOL {500.23 USD}')], expand=True), [
            '5.000 HOOL {500.23 USD}'
        ])
        self.assertEqual(self.render([I('5 HOOL {500.23 USD}, 12.3456 CAAD')], expand=True), [
            '12.3456 CAAD             ',
            ' 5.000  HOOL {500.23 USD}',
        ])

    def test_inventory_tabular(self):
        self.assertEqual(self.render([I('100 USD')], expand=False, listsep=' & '), [
            '100.00 USD'
        ])
        self.assertEqual(self.render([I('5 HOOL {500.23 USD}')], expand=False, listsep=' & '), [
            '5.000 HOOL {500.23 USD}'
        ])
        self.assertEqual(self.render([I('5 HOOL {500.23 USD}, 12.3456 CAAD')], expand=False, listsep=' & '), [
            '12.3456 CAAD & 5.000 HOOL {500.23 USD}',
        ])
        self.assertEqual(self.render([I('5 HOOL {500.23 USD}, 12.3456 CAAD'),
                                      I('55 HOOL {50.23 USD}, 2.3 CAAD')], expand=False, listsep=' & '), [
            '12.3456 CAAD &  5.000 HOOL {500.23 USD}',
            ' 2.3000 CAAD & 55.000 HOOL { 50.23 USD}',
        ])
        self.assertEqual(self.render([I('5 HOOL {500.23 USD}, 1 HOOL {567.89 USD}'),
                                      I('55 HOOL {50.23 USD}, 2.3 CAAD')], expand=False, listsep=' & '), [
            '         &  5.000 HOOL {500.23 USD} &  1.000 HOOL {567.89 USD}',
            '2.3 CAAD & 55.000 HOOL { 50.23 USD} &                         ',
        ])

    def test_inventory_too_many(self):
        self.assertEqual(self.render([I('10 AA, 2 BB, 3 CC, 4 DD'),
                                      I('5 AA, 6 EE, 7 FF')], expand=False, listsep=' & '), [
            '10 AA & 2 BB & 3 CC & 4 DD              ',
            ' 5 AA & 6 EE & 7 FF                     ',
        ])


class TestCostRenderer(RendererTestBase):

    def render(self, values, **kwargs):
        return super().render(Cost, values, **kwargs)

    def test_cost(self):
        self.dcontext.update(Decimal('1.0000'), 'ETH')
        self.assertEqual(self.render([Cost(D('1.0'), 'ETH', None, None),
                                      Cost(D('1.0'), 'ETH', datetime.date(2023, 8, 14), None),
                                      Cost(D('1.0'), 'ETH', datetime.date(2023, 8, 14), 'label')]), [
            '1.0000 ETH                     ',
            '1.0000 ETH, 2023-08-14         ',
            '1.0000 ETH, 2023-08-14, "label"',
        ])


class TestQueryRenderText(unittest.TestCase):

    def setUp(self):
        self.dcontext = display_context.DisplayContext()

    def render(self, types, rows, **kwargs):
        types = [Column(*t) for t in types]
        oss = io.StringIO()
        query_render.render_text(types, rows, self.dcontext, oss, **kwargs)
        return oss.getvalue()

    def test_render_simple(self):
        self.assertEqual(self.render(
            [('x', int), ('y', int), ('z', int)],
            [(1, 2, 3), (4, 5, 6)]), textwrap.dedent(
                """\
                x  y  z
                -  -  -
                1  2  3
                4  5  6
                """))

    def test_render_simple_unicode(self):
        self.assertEqual(self.render(
            [('x', int), ('y', int), ('z', int)],
            [(1, 2, 3), (4, 5, 6)], unicode=True), textwrap.dedent(
                """\
                x  y  z
                ─  ─  ─
                1  2  3
                4  5  6
                """))

    def test_render_boxed(self):
        self.assertEqual(self.render(
            [('x', int), ('y', int), ('z', int)],
            [(1, 2, 3), (4, 5, 6)], boxed=True), textwrap.dedent(
                """\
                +---+---+---+
                | x | y | z |
                +---+---+---+
                | 1 | 2 | 3 |
                | 4 | 5 | 6 |
                +---+---+---+
                """))

    def test_render_boxed_unicode(self):
        self.assertEqual(self.render(
            [('x', int), ('y', int), ('z', int)],
            [(1, 2, 3), (4, 5, 6)], boxed=True, unicode=True), textwrap.dedent(
                """\
                ┌───┬───┬───┐
                │ x │ y │ z │
                ├───┼───┼───┤
                │ 1 │ 2 │ 3 │
                │ 4 │ 5 │ 6 │
                └───┴───┴───┘
                """))

    def test_render_header_centering(self):
        self.assertEqual(self.render(
            [('x', int), ('y', int), ('z', int)],
            [(1, 22222, 3), (4, 5, 6)]), textwrap.dedent(
                """\
                x    y    z
                -  -----  -
                1  22222  3
                4      5  6
                """))

    def test_render_header_truncation(self):
        self.assertEqual(self.render(
            [('x', int), ('abcdefg', int), ('z', int)],
            [(1, 222, 3), (4, 5, 6)]), textwrap.dedent(
                """\
                x  abc  z
                -  ---  -
                1  222  3
                4    5  6
                """))

    def test_render_missing_values(self):
        self.assertEqual(self.render(
            [('xx', int), ('yy', int), ('zz', int)],
            [(12, None, 34), (None, 56, 78)]), textwrap.dedent(
                """\
                xx  yy  zz
                --  --  --
                12      34
                    56  78
                """))

    def test_render_missing_values_boxed(self):
        self.assertEqual(self.render(
            [('xx', int), ('yy', int), ('zz', int)],
            [(12, None, 34), (None, 56, 78)], boxed=True), textwrap.dedent(
                """\
                +----+----+----+
                | xx | yy | zz |
                +----+----+----+
                | 12 |    | 34 |
                |    | 56 | 78 |
                +----+----+----+
                """))

    def test_render_expand(self):
        self.assertEqual(self.render(
            [('x', int), ('inv', Inventory), ('q', int)],
            [(11, I('1.00 USD, 1.00 EUR, 1 TESTS'), 2),
             (33, I('2.00 EUR, 42 TESTS'), 4)], expand=True, boxed=True), textwrap.dedent(
                 """\
                 +----+-------------+---+
                 | x  |     inv     | q |
                 +----+-------------+---+
                 | 11 |  1.00 EUR   | 2 |
                 |    |  1    TESTS |   |
                 |    |  1.00 USD   |   |
                 | 33 |  2.00 EUR   | 4 |
                 |    | 42    TESTS |   |
                 +----+-------------+---+
                 """))

    def test_render_spaced(self):
        self.assertEqual(self.render(
            [('x', int)],
            [(1, ), (22, ), (333, )], spaced=True, boxed=True), textwrap.dedent(
                """\
                +-----+
                |  x  |
                +-----+
                |   1 |
                |     |
                |  22 |
                |     |
                | 333 |
                |     |
                +-----+
                """))


class TestQueryRenderCSV(unittest.TestCase):

    def setUp(self):
        self.dcontext = display_context.DisplayContext()

    def render(self, types, rows, **kwargs):
        oss = io.StringIO()
        query_render.render_csv(types, rows, self.dcontext, oss, **kwargs)
        # The csv modules emits DOS-style newlines.
        return oss.getvalue().replace('\r\n', '\n')

    def test_render_simple(self):
        self.assertEqual(self.render(
            [Column('x', int), Column('y', int), Column('z', int)],
            [(1, 2, 3), (4, 5, 6)]), textwrap.dedent(
                """\
                x,y,z
                1,2,3
                4,5,6
                """))

    def test_render_missing(self):
        self.assertEqual(self.render(
            [Column('x', int), Column('y', int), Column('z', int)],
            [(None, 2, 3), (4, None, 6)]), textwrap.dedent(
                """\
                x,y,z
                ,2,3
                4,,6
                """))

    def test_render_expand(self):
        self.assertEqual(self.render(
            [Column('x', int), Column('inv', Inventory), Column('q', int)],
            [(11, I('1.00 USD, 1.00 EUR, 1 TESTS'), 2),
             (33, I('2.00 EUR, 42 TESTS'), 4)], expand=True), "\n".join([
                 "x,inv,q",
                 "11, 1.00 EUR  ,2",
                 ", 1    TESTS,",
                 ", 1.00 USD  ,",
                 "33, 2.00 EUR  ,4",
                 ",42    TESTS,",
                 "",
             ]))



"""Rendering of rows.
"""
__copyright__ = "Copyright (C) 2014-2016  Martin Blais"
__license__ = "GNU GPLv2"

import collections
import csv
import datetime
import enum
import typing

from decimal import Decimal

from beancount.core import amount
from beancount.core import display_context
from beancount.core import inventory
from beancount.core import position


class Align(enum.Enum):
    LEFT = 0
    RIGHT = 1


class RenderContext:
    """Hold the query rendering configuration."""

    def __init__(self, dcontext, expand=False, listsep=', ', spaced=False, null=' '):
        self.dcontext = dcontext
        self.expand = expand
        self.listsep = listsep
        self.spaced = spaced
        self.null = null


# Map of data-types to renderer classes. This is populated by
# subclassing ColumnRenderer via an __init_subclass__ hook.
RENDERERS = {}


class ColumnRenderer:
    """Base class for classes that render column values.

    The column renderers are responsible to render uniform type values
    in a way that will align nicely in a column whereby all the values
    render to the same width.

    The formatters are instantiated and are feed all the values in the
    column via the ``update()`` method to accumulate the dimensions it
    will need to format them later on. The ``prepare()`` method then
    computes internal status required to format these values in
    consistent fashion. The ``width()`` method can then be used to
    retrieve the computer maximum width of the column. Individual
    values are formatted with the ``format()`` method. Values are
    assumed to be of the expected type for the formatter. Formatting
    values outside the set of the values fed via the ``update()``
    method is undefined behavior.

    """
    dtype = None
    align = Align.LEFT

    def __init__(self, ctx):
        self.maxwidth = 0
        self.prepared = False

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        RENDERERS[cls.dtype] = cls

    def update(self, value):
        """Update the rendered with the given value.

        Args:
          value: Any object of type ``dtype``.

        """

    def prepare(self):
        """Prepare to render the column.

        Returns:
          Computed column width.

        """
        self.prepared = True
        return self.maxwidth

    @property
    def width(self):
        if not self.prepared:
            raise RuntimeError('width property access before calling prepare()')
        return self.maxwidth

    def format(self, value):
        """Format the value.

        Args:
          value: Any object of type ``dtype``.

        Returns:
          A string or list of strings representing the rendered value.

        """
        raise NotImplementedError


class ObjectRenderer(ColumnRenderer):
    dtype = object

    def update(self, value):
        self.maxwidth = max(self.maxwidth, len(self.format(value)))

    def format(self, value):
        return str(value)


class DictRenderer(ObjectRenderer):
    dtype = dict


class BoolRenderer(ColumnRenderer):
    dtype = bool

    def update(self, value):
        self.maxwidth = max(self.maxwidth, 4 if value else 5)

    def format(self, value):
        return ('TRUE' if value else 'FALSE')


class StringRenderer(ObjectRenderer):
    dtype = str


class SetRenderer(ColumnRenderer):
    dtype = set

    def __init__(self, ctx):
        super().__init__(ctx)
        self.sep = ctx.listsep

    def update(self, value):
        self.maxwidth = max(self.maxwidth, sum(len(x) + len(self.sep) for x in value) - len(self.sep))

    def format(self, value):
        return self.sep.join(str(x) for x in sorted(value))


class DateRenderer(ColumnRenderer):
    dtype = datetime.date

    def update(self, value):
        self.maxwidth = 10

    def format(self, value):
        return value.strftime('%Y-%m-%d')


class IntRenderer(ObjectRenderer):
    dtype = int
    align = Align.RIGHT


class EnumRenderer(ObjectRenderer):
    dtype = enum.Enum

    def format(self, value):
        return value.name


class DecimalRenderer(ColumnRenderer):
    """Renderer for Decimal numbers.

    Numbers are left padded to align on the decimal point::

      -  123.40
      -    5.000
      -  -67

    """
    dtype = Decimal

    def __init__(self, ctx):
        super().__init__(ctx)
        # Max number of digits before the decimal point including sign.
        self.nintegral = 0
        # Max number of digits after the decimal point.
        self.nfractional = 0

    def update(self, value):
        n = value.as_tuple()
        if n.exponent > 0:
            # Special case for decimal numbers with positive exponent
            # and thus represented in scientific notation.
            self.nintegral = max(self.nintegral, len(str(value)))
        else:
            self.nintegral = max(self.nintegral, max(1, len(n.digits) + n.exponent) + n.sign)
            self.nfractional = max(self.nfractional, -n.exponent)

    def prepare(self):
        self.maxwidth = self.nintegral + self.nfractional + (1 if self.nfractional > 0 else 0)
        return super().prepare()

    def format(self, value):
        n = value.as_tuple()
        if n.exponent > 0:
            # Special case for decimal numbers with positive exponent
            # and thus represented in scientific notation.
            return str(value).rjust(self.nintegral).ljust(self.maxwidth)
        # Compute the padding required to align the decimal point.
        left = self.nintegral - (max(1, len(n.digits) + n.exponent) + n.sign)
        return f'{"":>{left}}{value:<{self.maxwidth - left}}'


class AmountRenderer(ColumnRenderer):
    """Renderer for Amount instances.

    The numerical part is formatted with the right quantization
    determined by ``dcontext`` in the rendering context and aligned on
    the decimal point across rows. Numbers are right padded with
    spaces to alignt the commodity symbols across rows::

      -  1234.00   USD
      -    42      TEST
      -     0.0001 ETH
      -   567.00   USD

    """
    dtype = amount.Amount

    def __init__(self, ctx):
        super().__init__(ctx)
        # Use the display context inferred from the input ledger to
        # determine the quantization of the column values.
        self.quantize = ctx.dcontext.quantize
        # Use column specific display context for formatting.
        self.dcontext = display_context.DisplayContext()
        # Maximum width of the commodity symbol.
        self.curwidth = 0

    def update(self, value):
        # Need to handle None to reuse this in PositionRenderer.
        if value is not None:
            number = self.quantize(value.number, value.currency)
            self.dcontext.update(number, value.currency)
            self.curwidth = max(self.curwidth, len(value.currency))

    def prepare(self):
        self.func = self.dcontext.build(display_context.Align.DOT)
        zero = Decimal()
        for commodity in self.dcontext.ccontexts:
            if commodity != '__default__':
                self.maxwidth = max(self.maxwidth, len(self.func(zero, commodity)) + 1 + self.curwidth)
        return super().prepare()

    def format(self, value):
        return f'{self.func(value.number, value.currency)} {value.currency:<{self.curwidth}}'


class CostRenderer(ObjectRenderer):
    dtype = position.Cost

    def __init__(self, ctx):
        super().__init__(ctx)
        self.amount_renderer = AmountRenderer(ctx)
        self.date_width = 0
        self.label_width = 0

    def update(self, value):
        self.amount_renderer.update(value)
        if value.date is not None:
            self.date_width = 10 + 2
        if value.label is not None:
            self.label_width = max(self.label_width, len(value.label) + 4)

    def prepare(self):
        cost_width = self.amount_renderer.prepare()
        self.maxwidth = cost_width + self.date_width + self.label_width
        return super().prepare()

    def format(self, value):
        parts = [self.amount_renderer.format(value)]
        if value.date is not None:
            parts.append(f'{value.date:%Y-%m-%d}')
        if value.label is not None:
            parts.append(f'"{value.label}"')
        return ', '.join(parts)


class PositionRenderer(ColumnRenderer):
    """Renderer for Position instrnaces.

    Both the unit numbers and the cost numbers are aligned::

      -    5.000 HOOL {500.23  USD }
      -  123     CA   {  1.000 HOOL}
      -    3.00  USD
      -   42.000 HOOL
      -    3.00  AAPL
      -    3.0   XY

    """
    dtype = position.Position

    def __init__(self, ctx):
        super().__init__(ctx)
        self.units_renderer = AmountRenderer(ctx)
        self.cost_renderer = AmountRenderer(ctx)

    def update(self, value):
        self.units_renderer.update(value.units)
        self.cost_renderer.update(value.cost)

    def prepare(self):
        units_width = self.units_renderer.prepare()
        cost_width = self.cost_renderer.prepare()
        self.maxwidth = units_width + cost_width + (3 if cost_width > 0 else 0)
        return super().prepare()

    def format(self, value):
        units = self.units_renderer.format(value.units)
        if value.cost is None:
            return units.ljust(self.maxwidth)
        cost = self.cost_renderer.format(value.cost)
        return f'{units} {{{cost}}}'


class InventoryRenderer(ColumnRenderer):
    """Renderer for Inventory instances.

    Inventories renders as a list of position strings. The format used
    differs whether expansion of list-like values to multiple rows in
    enabled or not.

    When row expansion is enabled, positions in each inventory values
    are sorted alphabetically by commodity symbol and are formatted
    with the same position formatter, resulting in all commodity
    strings to be aligned::

      -  1234.00   USD
           42      TEST
      -     0.0001 ETH
          567.00   USD

    When row expansion is disabled, the position formatters are unique
    for each commodity symbol and the values are rendered in a table
    like structure. The positions appear sorted by frequency of
    occurence in the column and alphabetically by commodity symbol::

      - 1234.00 USD 0.0001 ETH
      -  567.00 USD            42 TEST

    The separator between positions is determined by ``listsep`` in
    the rendering context.

    """
    dtype = inventory.Inventory

    def __init__(self, ctx):
        super().__init__(ctx)
        self.listsep = ctx.listsep
        # We look this up for each value, it makes sense to cache it
        # to avoid the attribute lookup in the context.
        self.expand = ctx.expand
        # How many times at most we have seen a commodity in an inventory.
        self.counts = collections.defaultdict(int)
        # Commodity specific renderers.
        self.renderers = collections.defaultdict(lambda: PositionRenderer(ctx))
        # How many distinct commodity need to be rendered.
        self.distinct = 0

    def update(self, value):
        for pos in value.get_positions():
            # We use the little indexing trick to do not have to
            # conditionalize this code on whether rows expansion is
            # enabled or not.
            self.renderers[self.expand or pos.units.currency].update(pos)
        counts = collections.Counter(pos.units.currency for pos in value.get_positions())
        for key, val in counts.items():
            self.counts[key] = max(self.counts[key], val)

    def prepare(self):
        if self.expand:
            self.maxwidth = self.renderers[self.expand].prepare()
        else:
            for commodity, renderer in self.renderers.items():
                w = renderer.prepare()
                self.distinct += self.counts[commodity]
                self.maxwidth += self.counts[commodity] * (w + len(self.listsep))
            self.maxwidth -= len(self.listsep)
        return super().prepare()

    @staticmethod
    def positionsortkey(position):
        # Sort positions combining fields in a more intuitive way than the default.
        return (position.units.currency, -position.units.number,
                (position.cost.currency, -position.cost.number, position.cost.date) if position.cost else ())

    def format(self, value):
        # Expanded row format.
        if self.expand:
            strings = []
            for pos in sorted(value.get_positions(), key=self.positionsortkey):
                strings.append(self.renderers[self.expand].format(pos))
            return strings
        # Too many distinct commodities to present in tabular format.
        if self.distinct > 5:
            strings = []
            for pos in sorted(value.get_positions(), key=self.positionsortkey):
                strings.append(self.renderers[pos.units.currency].format(pos))
            return self.listsep.join(strings).ljust(self.maxwidth)
        # Tabular format with same commodity positions vertically aligned.
        positions = collections.defaultdict(list)
        for pos in sorted(value.get_positions(), key=self.positionsortkey):
            positions[pos.units.currency].append(pos)
        strings = []
        for commodity, renderer in sorted(self.renderers.items()):
            strings += [renderer.format(pos) for pos in positions[commodity]]
            strings += [''.ljust(renderer.width)] * (self.counts[commodity] - len(positions[commodity]))
        return self.listsep.join(strings)


def render_rows(rows, renderers, ctx):
    """Render results set row."""

    # Filler for NULL values.
    null = ctx.null

    # Spacing row.
    spacerow = [''] * len(renderers)

    for row in rows:

        # Render the row cells. Do not pass missing values to the
        # renderers but substitute them with the appropriate
        # placeholder string.
        cells = [render.format(value) if value is not None else null for render, value in zip(renderers, row)]

        if not any(isinstance(cell, list) for cell in cells):
            # No multi line cells. Yield the row.
            yield cells

        else:
            # At least one multi line cell. Ensure that all cells are lists.
            cells = [cell if isinstance(cell, list) else [cell] for cell in cells]

            # Compute the maximum number of lines in any cell.
            nlines = max(len(cell) for cell in cells)

            # Add placeholder lines to short multi line cells.
            for cell in cells:
                if len(cell) < nlines:
                    cell.extend([''] * (nlines - len(cell)))

            # Yield the obtained rows.
            yield from zip(*cells)

        # Add spacing row when needed.
        if ctx.spaced:
            yield spacerow


def _get_renderer(datatype, ctx):
    datatype = typing.get_origin(datatype) or datatype
    for d in datatype.__mro__:  # pragma: no branch
        renderer = RENDERERS.get(d)
        if renderer:
            return renderer(ctx)


def render_text(columns, rows, dcontext, file, expand=False, boxed=False,
                spaced=False, listsep='  ', nullvalue='', narrow=True, unicode=False, **kwargs):
    """Render the result of executing a query in text format.

    Args:
      columns: A list of beanquery.Column descrining the table columns.
      rows: Data to render.
      dcontext: A DisplayContext object prepared for rendering numbers.
      file: A file object to render the results to.
      expand: When true expand columns that render to lists to multiple rows.
      boxed: When true draw an ascii-art table borders.
      spaced: When true insert an empty line between rows.
      listsep: String to use to separate values in list-like column values.
      nullvalue: String to use to represent NULL values.
      narrow: When true truncate headers to the maximum column values width.
      unicode: When true use unicode box drawing characters to draw tables.

    """
    ctx = RenderContext(dcontext, expand=expand, spaced=spaced, listsep=listsep, null=nullvalue)
    renderers = [_get_renderer(column.datatype, ctx) for column in columns]
    headers = [column.name for column in columns]
    alignment = [renderer.align for renderer in renderers]

    # Prime the renderers.
    for row in rows:
        for value, renderer in zip(row, renderers):
            if value is not None:
                renderer.update(value)

    # Compute columns widths.
    widths = [max(1, narrow or len(header), len(nullvalue), render.prepare()) for header, render in zip(headers, renderers)]

    # Initialize table style. For unicode box drawing characters,
    # see https://www.unicode.org/charts/PDF/U2500.pdf
    if boxed:
        if unicode:
            frmt = '\u2502 {} \u2502\n'
            colsep = ' \u2502 '
            lines = [''.rjust(width, '\u2500') for width in widths]
            top =    '\u250C\u2500{}\u2500\u2510\n'.format('\u2500\u252C\u2500'.join(lines))
            hline =  '\u251C\u2500{}\u2500\u2524\n'.format('\u2500\u253C\u2500'.join(lines))
            bottom = '\u2514\u2500{}\u2500\u2518\n'.format('\u2500\u2534\u2500'.join(lines))
        else:
            frmt = '| {} |\n'
            colsep = ' | '
            top = hline = bottom = '+-{}-+\n'.format('-+-'.join(''.rjust(width, '-') for width in widths))
    else:
        frmt = '{}\n'
        colsep = '  '
        top = bottom = ''
        hline = '{}\n'.format(colsep.join(''.rjust(width, '\u2500' if unicode else '-') for width in widths))

    # Header.
    file.write(top)
    file.write(frmt.format(colsep.join(header[:width].center(width) for header, width in zip(headers, widths))))
    file.write(hline)

    # Rows.
    for row in render_rows(rows, renderers, ctx):
        file.write(frmt.format(colsep.join(x.ljust(w) if a == Align.LEFT else x.rjust(w)
                                           for x, w, a in zip(row, widths, alignment))))

    # Footer.
    file.write(bottom)


def render_csv(columns, rows, dcontext, file, expand=False, nullvalue='', **kwargs):
    """Render the result of executing a query in text format.

    Args:
      columns: A list of beanquery.Column describing the table columns.
      rows: Data to render.
      dcontext: A DisplayContext object prepared for rendering numbers.
      file: A file object to render the results to.
      expand: A boolean, if true, expand columns that render to lists on multiple rows.
      nullvalue: String to use to represent NULL values.
    """
    ctx = RenderContext(dcontext, expand=expand, spaced=False, listsep=',', null=nullvalue)
    renderers = [_get_renderer(column.datatype, ctx) for column in columns]
    headers = [column.name for column in columns]

    # Prime the renderers.
    for row in rows:
        for value, renderer in zip(row, renderers):
            if value is not None:
                renderer.update(value)

    # Prepare the renders.
    [render.prepare() for render in renderers]

    # Write the CSV file.
    writer = csv.writer(file)
    writer.writerow(headers)
    writer.writerows(render_rows(rows, renderers, ctx))



import csv
import datetime

from os import path
from urllib.parse import urlparse, parse_qsl

from beanquery import tables
from beanquery import query_compile
from beanquery.parser import BQLParser, BQLSemantics

# Support conversions from all fundamental types supported by the BQL parser:
#
#   literal
#       =
#       | date
#       | decimal
#       | integer
#       | string
#       | null
#       | boolean
#       ;
#

def _guess_type(value):
    try:
        r = BQLParser().parse(value, start='literal', semantics=BQLSemantics())
    except Exception:
        # Everything that is not recognized as something else is a string.
        return str
    return type(r)


def _parse_bool(value):
    x = value.strip().lower()
    if x == '1' or x == 'true':
        return True
    if x == '0' or x == 'false':
        return False
    raise ValueError(value)


_TYPES_TO_PARSERS = {
    bool: _parse_bool,
    datetime.date: datetime.date.fromisoformat,
}


class Column(query_compile.EvalColumn):
    def __init__(self, key, datatype, func):
        super().__init__(datatype)
        self.key = key
        self.func = func

    def __call__(self, row):
        return self.func(row[self.key])


class Table(tables.Table):
    def __init__(self, name, columns, data, header=False, **fmtparams):
        self.name = name
        self.data = data
        self.header = header
        # Skip white space after field separator by default to make parsing
        # columns accordingly to their type easier, unless the setting is
        # overridden by the user.
        fmtparams.setdefault('skipinitialspace', True)
        self.reader = csv.reader(data, **fmtparams)
        self.columns = {}
        if columns is None:
            names = next(self.reader, [])
            values = next(self.reader, [])
            datatypes = (_guess_type(value) for value in values)
            columns = zip(names, datatypes)
        for cname, ctype in columns:
            converter = _TYPES_TO_PARSERS.get(ctype, ctype)
            self.columns[cname] = Column(len(self.columns), ctype, converter)

    def __del__(self):
        self.data.close()

    def __iter__(self):
        self.data.seek(0)
        it = iter(self.reader)
        if self.header:
            next(it)
        return it


def create(name, columns, using):
    parts = urlparse(using)
    filename = parts.path
    params = dict(parse_qsl(parts.query))
    encoding = params.pop('encoding', None)
    header = params.pop('header', columns is None)
    if filename:
        data = open(filename, encoding=encoding)
    return Table(name, columns, data, header=header, **params)


def attach(context, dsn, *, data=None):
    parts = urlparse(dsn)
    filename = parts.path
    params = dict(parse_qsl(parts.query))
    encoding = params.pop('encoding', None)
    if filename:
        data = open(filename, encoding=encoding)
    name = params.pop('name', None) or path.splitext(path.basename(filename))[0] or 'csv'
    context.tables[name] = Table(name, None, data, header=True, **params)



from beancount.core import display_context
from beancount.parser import printer


def render(desc, rows, file, *, dcontext, **kwargs):
    # Create a display context that renders all numbers with their
    # natural precision, but honors the commas option in the ledger.
    commas = dcontext.commas
    dcontext = display_context.DisplayContext()
    dcontext.set_commas(commas)
    return printer.print_entries([entry for entry, in rows], dcontext, file=file)



import datetime
import decimal

import tatsu

from ..errors import ProgrammingError
from .parser import BQLParser
from . import ast


class BQLSemantics:

    def set_context(self, ctx):
        self._ctx = ctx

    def null(self, value):
        return None

    def integer(self, value):
        return int(value)

    def decimal(self, value):
        return decimal.Decimal(value)

    def date(self, value):
        return datetime.date.fromisoformat(value)

    def string(self, value):
        return value[1:-1]

    def boolean(self, value):
        return value == 'TRUE'

    def unquoted_identifier(self, value):
        return value.lower()

    def quoted_identifier(self, value):
        return value.replace('""', '"')

    def asterisk(self, value):
        return ast.Asterisk()

    def list(self, value):
        return list(value)

    def ordering(self, value):
        return ast.Ordering[value or 'ASC']

    def _default(self, value, typename=None):
        if typename is not None:
            func = getattr(ast, typename)
            return func(**{name.rstrip('_'): value for name, value in value.items()})
        return value


class ParseError(ProgrammingError):
    def __init__(self, parseinfo):
        super().__init__('syntax error')
        self.parseinfo = parseinfo


def parse(text):
    try:
        return BQLParser().parse(text, semantics=BQLSemantics())
    except tatsu.exceptions.ParseError as exc:
        line = exc.tokenizer.line_info(exc.pos).line
        parseinfo = tatsu.infos.ParseInfo(exc.tokenizer, exc.item, exc.pos, exc.pos + 1, line, [])
        raise ParseError(parseinfo) from exc



"""Code to split table columns containing amounts and inventories into number columns.

For example, given a column with this content:

    ----- amount ------
    101.23 USD
       200 JPY
     99.23 USD
     38.34 USD, 100 JPY

We can convert this into two columns and remove the currencies:

    -amount (USD)- -amount (JPY)-
            101.23
                              200
             99.23
             38.34            100

The point is that the columns should be typed as numbers to make this importable
into a spreadsheet and able to be processed.

Notes:

* This handles the Amount, Position and Inventory datatypes. There is code to
  automatically recognize columns containing such types from a table of strings
  and convert such columns to their corresponding guessed data types.

* The per-currency columns are ordered in decreasing order of the number of
  instances of numbers seen for each currency. So if the most numbers you have
  in a column are USD, then the USD column renders first.

* Cost basis specifications should be unmodified and reported to a dedicated
  extra column, like this:

    ----- amount ------
    1 AAPL {21.23 USD}

  We can convert this into two columns and remove the currencies:

    -amount (AAPL)- -Cost basis-
                  1 {21.23 USD}

  (Eventually we might support the conversion of cost amounts as well, but they
  may contain other information, such as a label or a date, so for now we don't
  convert them. I'm not sure there's a good practical use case in doing that
  yet.)

* We may provide some options to break out only some of the currencies into
  columns, in order to handle the case where an inventory contains a large
  number of currencies and we want to only operate on a restricted set of
  operating currencies.

* If you provide a DisplayFormatter object to the numberification routine, they
  quantize each column according to their currency's precision. It is
  recommended that you do that.

"""
__copyright__ = "Copyright (C) 2015-2017  Martin Blais"
__license__ = "GNU GPLv2"

import collections
from decimal import Decimal

from beancount.core import amount
from beancount.core import position
from beancount.core import inventory

from .cursor import Column


def numberify_results(columns, drows, dformat=None):
    """Number rows containing Amount, Position or Inventory types.

    Args:
      result_types: A list of items describing the names and data types of the items in
        each column.
      result_rows: A list of ResultRow instances.
      dformat: An optional DisplayFormatter. If set, quantize the numbers by
        their currency-specific precision when converting the Amount's,
        Position's or Inventory'es..
    Returns:
      A pair of modified (result_types, result_rows) with converted datatypes.
    """
    # Build an array of converters.
    converters = []
    for index, column in enumerate(columns):
        convert_col_fun = CONVERTING_TYPES.get(column.datatype)
        if convert_col_fun is None:
            converters.append(IdentityConverter(column.name, column.datatype, index))
        else:
            col_converters = convert_col_fun(column.name, drows, index)
            converters.extend(col_converters)

    # Derive the output types from the expected outputs from the converters
    # themselves.
    otypes = tuple(Column(c.name, c.dtype) for c in converters)

    # Convert the input rows by processing them through the converters.
    orows = []
    for drow in drows:
        orow = []
        for converter in converters:
            orow.append(converter(drow, dformat))
        orows.append(orow)

    return otypes, orows


class IdentityConverter:
    """A converter that simply copies its column."""

    def __init__(self, name, dtype, index):
        self.name = name
        self.dtype = dtype
        self.index = index

    def __call__(self, drow, _):
        return drow[self.index]


class AmountConverter:
    """A converter that extracts the number of an amount for a specific currency."""

    dtype = Decimal

    def __init__(self, name, index, currency):
        self.name = name
        self.index = index
        self.currency = currency

    def __call__(self, drow, dformat):
        vamount = drow[self.index]
        if vamount and vamount.currency == self.currency:
            number = vamount.number
            if dformat:
                number = dformat.quantize(number, self.currency)
        else:
            number = None
        return number


def convert_col_Amount(name, drows, index):
    """Create converters for a column of type Amount.

    Args:
      name: A string, the column name.
      drows: The table of objects.
      index: The column number.
    Returns:
      A list of Converter instances, one for each of the currency types found.
    """
    currency_map = collections.defaultdict(int)
    for drow in drows:
        vamount = drow[index]
        if vamount and vamount.currency:
            currency_map[vamount.currency] += 1
    return [AmountConverter('{} ({})'.format(name, currency), index, currency)
            for currency, _ in sorted(currency_map.items(),
                                      key=lambda item: (item[1], item[0]),
                                      reverse=True)]


class PositionConverter:
    """A converter that extracts the number of a position for a specific currency."""

    dtype = Decimal

    def __init__(self, name, index, currency):
        self.name = name
        self.index = index
        self.currency = currency

    def __call__(self, drow, dformat):
        pos = drow[self.index]
        if pos and pos.units.currency == self.currency:
            number = pos.units.number
            if dformat:
                number = dformat.quantize(pos.units.number, self.currency)
        else:
            number = None
        return number


def convert_col_Position(name, drows, index):
    """Create converters for a column of type Position.

    Args:
      name: A string, the column name.
      drows: The table of objects.
      index: The column number.
    Returns:
      A list of Converter instances, one for each of the currency types found.
    """
    currency_map = collections.defaultdict(int)
    for drow in drows:
        pos = drow[index]
        if pos and pos.units.currency:
            currency_map[pos.units.currency] += 1
    return [PositionConverter('{} ({})'.format(name, currency), index, currency)
            for currency, _ in sorted(currency_map.items(),
                                      key=lambda item: (item[1], item[0]),
                                      reverse=True)]


class InventoryConverter:
    """A converter that extracts the number of a inventory for a specific currency.
    If there are multiple lots we aggregate by currency."""

    dtype = Decimal

    def __init__(self, name, index, currency):
        self.name = name
        self.index = index
        self.currency = currency

    def __call__(self, drow, dformat):
        inv = drow[self.index]
        # FIXME:: get_currency_units() returns ZERO and not None when the value
        # isn't present. This should be fixed to distinguish between the two.
        number = inv.get_currency_units(self.currency).number
        if number and dformat:
            number = dformat.quantize(number, self.currency)
        return number or None


def convert_col_Inventory(name, drows, index):
    """Create converters for a column of type Inventory.

    Args:
      name: A string, the column name.
      drows: The table of objects.
      index: The column number.
    Returns:
      A list of Converter instances, one for each of the currency types found.
    """
    currency_map = collections.defaultdict(int)
    for drow in drows:
        inv = drow[index]
        for currency in inv.currencies():
            currency_map[currency] += 1
    return [InventoryConverter('{} ({})'.format(name, currency), index, currency)
            for currency, _ in sorted(currency_map.items(),
                                      key=lambda item: (item[1], item[0]),
                                      reverse=True)]


# A mapping of data types to their converter factory.
CONVERTING_TYPES = {
    amount.Amount       : convert_col_Amount,
    position.Position   : convert_col_Position,
    inventory.Inventory : convert_col_Inventory,
}



__copyright__ = "Copyright (C) 2014-2016  Martin Blais"
__license__ = "GNU GPLv2"

import datetime
import unittest

from decimal import Decimal as D

from beancount import loader

import beanquery

from beanquery import Connection, CompilationError, ProgrammingError
from beanquery import compiler
from beanquery import query_compile as qc
from beanquery import query_env as qe
from beanquery import parser
from beanquery import tables
from beanquery.parser import ast
from beanquery.sources import test


class Table:
    # mock table to be used in tests
    def __init__(self, name):
        self.name = name

    def __eq__(self, other):
        if not isinstance(other, tables.Table):
            return NotImplemented
        return other.name == self.name


class Column(qc.EvalColumn):
    # mock column to be used in tests
    def __init__(self, name, dtype=str):
        self.name = name
        self.dtype = dtype

    def __eq__(self, other):
        if not isinstance(other, qc.EvalColumn):
            return False
        return getattr(other, 'name', type(other).__name__) == self.name


class TestCompileExpression(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.context = Connection()
        cls.context.tables['test'] = test.Table(0)
        # parser for expressions
        cls.parser = parser.BQLParser(start='expression')
        cls.compiler = compiler.Compiler(cls.context)
        cls.compiler.table = cls.context.tables['test']

    def compile(self, expr):
        expr = self.parser.parse(expr, semantics=parser.BQLSemantics())
        return self.compiler.compile(expr)

    def test_expr_invalid(self):
        with self.assertRaises(CompilationError):
            self.compile('''invalid''')

    def test_expr_column(self):
        self.assertEqual(self.compile('''x'''), Column('x', int))

    def test_expr_function(self):
        self.assertEqual(self.compile('''sum(x)'''), qe.SumInt(None, [Column('x', int)]))

    def test_expr_unaryop(self):
        self.assertEqual(self.compile('''not x'''), qc.Operator(ast.Not, [Column('x', int)]))

    def test_expr_binaryop(self):
        self.assertEqual(self.compile('''x = 1'''), qc.Operator(ast.Equal, [Column('x', int), qc.EvalConstant(1)]))

    def test_expr_constant(self):
        self.assertEqual(self.compile('''17'''), qc.EvalConstant(D('17')))

    def test_expr_function_arity(self):
        # compile with an incorrect number of arguments.
        with self.assertRaises(CompilationError):
            self.compile('''sum(1, 2)''')

    def test_constants_folding(self):
        # unary op
        self.assertEqual(self.compile('''-2'''), qc.EvalConstant(D('-2')))
        # binary op
        self.assertEqual(self.compile('''2 + 2'''), qc.EvalConstant(D('4')))
        # funtion
        self.assertEqual(self.compile('''root('Assets:Cash', 1)'''), qc.EvalConstant('Assets'))


class TestCompileAggregateChecks(unittest.TestCase):

    def test_is_aggregate_derived(self):
        columns, aggregates = compiler.get_columns_and_aggregates(
            qc.EvalAnd([
                qc.Operator(ast.Equal, [
                    Column('lineno', int),
                    qc.EvalConstant(42),
                ]),
                qc.EvalOr([
                    qc.Operator(ast.Not, [
                        qc.Operator(ast.Equal, [
                            Column('date', datetime.date),
                            qc.EvalConstant(datetime.date(2014, 1, 1)),
                        ]),
                    ]),
                    qc.EvalConstant(False),
                ]),
            ]))
        self.assertEqual((2, 0), (len(columns), len(aggregates)))

        columns, aggregates = compiler.get_columns_and_aggregates(
            qc.EvalAnd([
                qc.Operator(ast.Equal, [
                    Column('lineno', int),
                    qc.EvalConstant(42),
                ]),
                qc.EvalOr([
                    qc.Operator(ast.Not, [
                        qc.Operator(ast.Not, [
                            qc.Operator(ast.Equal, [
                                Column('date', datetime.date),
                                qc.EvalConstant(datetime.date(2014, 1, 1)),
                            ]),
                        ]),
                    ]),
                    # Aggregation node deep in the tree.
                    qe.SumInt(None, [qc.EvalConstant(1)]),
                ]),
            ]))
        self.assertEqual((2, 1), (len(columns), len(aggregates)))

    def test_get_columns_and_aggregates(self):
        # Simple column.
        c_query = Column('position')
        columns, aggregates = compiler.get_columns_and_aggregates(c_query)
        self.assertEqual((1, 0), (len(columns), len(aggregates)))
        self.assertFalse(compiler.is_aggregate(c_query))

        # Multiple columns.
        c_query = qc.EvalAnd([Column('position'), Column('date')])
        columns, aggregates = compiler.get_columns_and_aggregates(c_query)
        self.assertEqual((2, 0), (len(columns), len(aggregates)))
        self.assertFalse(compiler.is_aggregate(c_query))

        # Simple aggregate.
        c_query = qe.SumPosition(None, [Column('position')])
        columns, aggregates = compiler.get_columns_and_aggregates(c_query)
        self.assertEqual((0, 1), (len(columns), len(aggregates)))
        self.assertTrue(compiler.is_aggregate(c_query))

        # Multiple aggregates.
        c_query = qc.EvalAnd([qe.First(None, [Column('date')]), qe.Last(None, [Column('flag')])])
        columns, aggregates = compiler.get_columns_and_aggregates(c_query)
        self.assertEqual((0, 2), (len(columns), len(aggregates)))
        self.assertTrue(compiler.is_aggregate(c_query))

        # Simple non-aggregate function.
        c_query = qe.Function('length', [Column('account')])
        columns, aggregates = compiler.get_columns_and_aggregates(c_query)
        self.assertEqual((1, 0), (len(columns), len(aggregates)))
        self.assertFalse(compiler.is_aggregate(c_query))

        # Mix of column and aggregates (this is used to detect this illegal case).
        c_query = qc.EvalAnd([
            qe.Function('length', [Column('account')]),
            qe.SumPosition(None, [Column('position')]),
        ])
        columns, aggregates = compiler.get_columns_and_aggregates(c_query)
        self.assertEqual((1, 1), (len(columns), len(aggregates)))
        self.assertTrue(compiler.is_aggregate(c_query))


class CompileSelectBase(unittest.TestCase):

    maxDiff = 8192

    def setUp(self):
        entries, errors, options = loader.load_string('')
        self.ctx = Connection('beancount:', entries=entries, errors=errors, options=options)

    def compile(self, query):
        """Parse one query and compile it.

        Args:
          query: An SQL query to be parsed.
        Returns:
          The AST.
        """
        c_query = self.ctx.compile(self.ctx.parse(query))
        if isinstance(c_query, ast.Select):
            self.assertSelectInvariants(c_query)
        return c_query

    def assertSelectInvariants(self, query):
        """Assert the invariants on the query.

        Args:
          query: An instance of EvalQuery, a compiled query statement.
        Raises:
          AssertionError: if the check fails.
        """
        # Check that the group references cover all the simple indexes.
        if query.group_indexes is not None:
            non_aggregate_indexes = [index
                                     for index, c_target in enumerate(query.c_targets)
                                     if not compiler.is_aggregate(c_target.c_expr)]

            self.assertEqual(set(non_aggregate_indexes), set(query.group_indexes),
                             "Invalid indexes: {}".format(query))

    def assertIndexes(self,
                      query,
                      expected_simple_indexes,
                      expected_aggregate_indexes,
                      expected_group_indexes,
                      expected_order_spec):
        """Check the four lists of indexes for comparison.

        Args:
          query: An instance of EvalQuery, a compiled query statement.
          expected_simple_indexes: The expected visible non-aggregate indexes.
          expected_aggregate_indexes: The expected visible aggregate indexes.
          expected_group_indexes: The expected group_indexes.
          expected_order_spec: The expected order_spec.
        Raises:
          AssertionError: if the check fails.
        """
        # Compute the list of _visible_ aggregates and non-aggregates.
        simple_indexes = [index
                          for index, c_target in enumerate(query.c_targets)
                          if c_target.name and not compiler.is_aggregate(c_target.expression)]
        aggregate_indexes = [index
                             for index, c_target in enumerate(query.c_targets)
                             if c_target.name and compiler.is_aggregate(c_target.expression)]

        self.assertEqual(set(expected_simple_indexes), set(simple_indexes))

        self.assertEqual(set(expected_aggregate_indexes), set(aggregate_indexes))

        self.assertEqual(
            set(expected_group_indexes) if expected_group_indexes is not None else None,
            set(query.group_indexes) if query.group_indexes is not None else None)

        self.assertEqual(
            set(expected_order_spec) if expected_order_spec is not None else None,
            set(query.order_spec) if query.order_spec is not None else None)

    def assertCompile(self, expected, query, debug=False):
        """Assert parsed and compiled contents from 'query' is 'expected'.

        Args:
          expected: An expected AST to compare against the parsed value.
          query: An SQL query to be parsed.
          debug: A boolean, if true, print extra debugging information on the console.
        Raises:
          AssertionError: If the actual AST does not match the expected one.
        """
        actual = self.compile(query)
        if debug:
            print()
            print()
            print(actual)
            print()
        try:
            self.assertEqual(expected, actual)
            return actual
        except AssertionError:
            print()
            print("Expected: {}".format(expected))
            print("Actual  : {}".format(actual))
            raise


class TestCompileSelect(CompileSelectBase):

    def test_compile_from(self):
        # Test the compilation of from.

        query = self.compile("SELECT account FROM CLOSE;")
        self.assertEqual(query.table.close, True)

        query = self.compile("SELECT account FROM length(payee) != 0;")
        self.assertTrue(isinstance(query.c_where, qc.EvalNode))

        with self.assertRaises(CompilationError):
            query = self.compile("SELECT account FROM sum(payee) != 0;")

    def test_compile_from_invalid_dates(self):
        self.compile("""
          SELECT account FROM  OPEN ON 2014-03-01  CLOSE ON 2014-03-02;
        """)

        self.compile("""
          SELECT account FROM  OPEN ON 2014-03-02  CLOSE ON 2014-03-02;
        """)

        with self.assertRaises(CompilationError):
            self.compile("""
              SELECT account FROM  OPEN ON 2014-03-03  CLOSE ON 2014-03-02;
            """)

    def test_compile_targets_wildcard(self):
        # Test the wildcard expansion.
        query = self.compile("SELECT *;")
        self.assertTrue(list, type(query.c_targets))
        self.assertGreater(len(query.c_targets), 3)
        self.assertTrue(all(isinstance(target.c_expr, qc.EvalColumn) for target in query.c_targets))

    def test_compile_targets_named(self):
        # Test the wildcard expansion.
        query = self.compile("SELECT length(account), account as a, date;")
        self.assertEqual(
            [qc.EvalTarget(qe.Function('length', [Column('account')]), 'length(account)', False),
             qc.EvalTarget(Column('account'), 'a', False),
             qc.EvalTarget(Column('date'), 'date', False)],
            query.c_targets)

    def test_compile_mixed_aggregates(self):
        # Check mixed aggregates and non-aggregates in a target.
        with self.assertRaises(CompilationError) as assertion:
            self.compile("""
              SELECT length(account) and sum(length(account));
            """)
        self.assertRegex(str(assertion.exception), 'mixed aggregates and non-aggregates')

    def test_compile_aggregates_of_aggregates(self):
        # Check mixed aggregates and non-aggregates in a target.
        with self.assertRaises(CompilationError) as assertion:
            self.compile("""
              SELECT sum(sum(length(account)));
            """)
        self.assertRegex(str(assertion.exception), 'aggregates of aggregates')

    def test_compile_having_non_aggregate(self):
        with self.assertRaises(CompilationError) as assertion:
            self.compile("""
              SELECT account, sum(number) GROUP BY account HAVING flag;
            """)
        self.assertRegex(str(assertion.exception), 'the HAVING clause must be an aggregate')

    def test_compile_group_by_inventory(self):
        with self.assertRaises(CompilationError):
            self.compile("""
              SELECT sum(number), balance GROUP BY balance;
            """)


class TestCompileSelectGroupBy(CompileSelectBase):

    def test_compile_group_by_non_aggregates(self):
        self.compile("""
          SELECT payee GROUP BY payee, length(account);
        """)

        with self.assertRaises(CompilationError) as assertion:
            self.compile("""
              SELECT payee GROUP BY payee, last(account);
            """)
        self.assertRegex(str(assertion.exception), 'may not be aggregates')

    def test_compile_group_by_reference_by_name(self):
        # Valid references to target names.
        self.compile("""
          SELECT payee, last(account) GROUP BY payee;
        """)
        self.compile("""
          SELECT payee as a, last(account) as len GROUP BY a;
        """)

        # References to non-targets have to be valid.
        self.compile("""
          SELECT payee, last(account) as len GROUP BY payee, date;
        """)

        with self.assertRaises(CompilationError):
            self.compile("""
              SELECT payee, last(account) as len GROUP BY something;
            """)

    def test_compile_group_by_reference_by_number(self):
        self.compile("""
          SELECT date, payee, narration GROUP BY 1, 2, 3;
        """)

        with self.assertRaises(CompilationError):
            self.compile("""
              SELECT date, payee, narration GROUP BY 1, 2, 3, 4;
            """)

    def test_compile_group_by_reference_an_aggregate(self):
        # By name.
        with self.assertRaises(CompilationError):
            self.compile("""
              SELECT payee, last(account) as last GROUP BY last;
            """)
        with self.assertRaises(CompilationError):
            self.compile("""
              SELECT account, sum(number) as sum_num GROUP BY account, sum_num;
            """)

        # By number.
        with self.assertRaises(CompilationError):
            self.compile("""
              SELECT payee, last(account) as last GROUP BY 2;
            """)

        # Explicit aggregate in group-by clause.
        with self.assertRaises(CompilationError):
            self.compile("""
              SELECT account, sum(number) GROUP BY account, sum(number);
            """)

    def test_compile_group_by_implicit(self):
        self.compile("""
          SELECT payee, last(account);
        """)

        self.compile("""
          SELECT first(account), last(account);
        """)

    def test_compile_group_by_coverage(self):
        # Non-aggregates.
        query = self.compile("SELECT account, length(account);")
        self.assertEqual(None, query.group_indexes)
        self.assertEqual(None, query.order_spec)

        # Aggregates only.
        query = self.compile("SELECT first(account), last(account);")
        self.assertEqual([], query.group_indexes)

        # Mixed with non-aggregates in group-by clause.
        query = self.compile("SELECT account, sum(number) GROUP BY account;")
        self.assertEqual([0], query.group_indexes)

        # Mixed with non-aggregates in group-by clause with non-aggregates a
        # strict subset of the group-by columns. 'account' is a subset of
        # {'account', 'flag'}.
        query = self.compile("""
          SELECT account, sum(number) GROUP BY account, flag;
        """)
        self.assertEqual([0, 2], query.group_indexes)

        # Non-aggregates not covered by group-by clause.
        with self.assertRaises(CompilationError):
            self.compile("""
              SELECT account, date, sum(number) GROUP BY account;
            """)
        with self.assertRaises(CompilationError):
            self.compile("""
              SELECT payee, last(account) as len GROUP BY date;
            """)

        # Non-aggregates not covered by group-by clause, and no aggregates in
        # the list of targets.
        with self.assertRaises(CompilationError):
            self.compile("""
              SELECT date, flag, account, number GROUP BY date, flag;
            """)

        # All non-aggregates and matching list of aggregates (this is a
        # pointless list of aggregates, essentially).
        query = self.compile("""
          SELECT date, flag, account GROUP BY date, flag, account;
        """)
        self.assertEqual([0, 1, 2], query.group_indexes)

    def test_compile_group_by_reconcile(self):
        # Check that no invisible column is created if redundant.
        query = self.compile("""
          SELECT account, length(account), sum(number)
          GROUP BY account, length(account);
        """)
        self.assertEqual([0, 1], query.group_indexes)


class TestCompileSelectOrderBy(CompileSelectBase):

    def test_compile_order_by_simple(self):
        query = self.compile("""
          SELECT account, sum(number) GROUP BY account ORDER BY account;
        """)
        self.assertEqual([0], query.group_indexes)
        self.assertEqual([(0, False)], query.order_spec)

    def test_compile_order_by_simple_2(self):
        query = self.compile("""
          SELECT account, length(narration) GROUP BY account, 2 ORDER BY 1, 2;
        """)
        self.assertEqual([0, 1], query.group_indexes)
        self.assertEqual([(0, False), (1, False)], query.order_spec)

        query = self.compile("""
          SELECT account, length(narration) as l GROUP BY account, l ORDER BY l;
        """)
        self.assertEqual([0, 1], query.group_indexes)
        self.assertEqual([(1, False)], query.order_spec)

    def test_compile_order_by_create_non_agg(self):
        with self.assertRaises(CompilationError):
            self.compile("""
              SELECT account, last(narration) GROUP BY account ORDER BY year(date);
            """)

        with self.assertRaises(CompilationError):
            self.compile("""
              SELECT account GROUP BY account ORDER BY year(date);
            """)

        query = self.compile("""
          SELECT account, year(date) GROUP BY 1, 2 ORDER BY 2;
        """)
        self.assertEqual([0, 1], query.group_indexes)
        self.assertEqual([(1, False)], query.order_spec)

        # We detect similarity between order-by and targets yet.
        self.compile("""
          SELECT account, year(date) GROUP BY 1, 2 ORDER BY year(date);
        """)

    def test_compile_order_by_reconcile(self):
        # Check that no invisible column is created if redundant.
        query = self.compile("""
          SELECT account, length(account)
          ORDER BY length(account);
        """)
        self.assertEqual([(1, False)], query.order_spec)

    def test_compile_order_by_reference_invisible(self):
        # So this is an interesting case: the grouping expression is an
        # invisible non-aggregate (length(account)) and the ordering expression
        # refers to the same non-aggregate expression. If they are reconciled to
        # the same invisible expression, the condition that the grouping
        # expressions cover all the non-aggregates is fulfilled. Otherwise, it
        # would fail. In order to support the compilation of this, we must
        # reconcile the grouping and ordering columns by comparing their values.
        query = self.compile("""
          SELECT count(account) as num, first(account) as first
          GROUP BY length(account)
          ORDER BY length(account);
        """)
        self.assertEqual([2], query.group_indexes)
        self.assertEqual([(2, False)], query.order_spec)

    def test_compile_order_by_aggregate(self):
        query = self.compile("""
          SELECT account, first(narration) GROUP BY account ORDER BY 2;
        """)
        self.assertEqual([0], query.group_indexes)
        self.assertEqual([(1, False)], query.order_spec)

        query = self.compile("""
          SELECT account, first(narration) as f GROUP BY account ORDER BY f;
        """)
        self.assertEqual([0], query.group_indexes)
        self.assertEqual([(1, False)], query.order_spec)

        query = self.compile("""
          SELECT account, first(narration) GROUP BY account ORDER BY sum(number);
        """)
        self.assertEqual([0], query.group_indexes)
        self.assertEqual([(2, False)], query.order_spec)

        query = self.compile("""
          SELECT account GROUP BY account ORDER BY sum(number);
        """)
        self.assertEqual([0], query.group_indexes)
        self.assertEqual([(1, False)], query.order_spec)


class TestTranslationJournal(CompileSelectBase):

    maxDiff = 4096

    def test_journal(self):
        journal = parser.parse("JOURNAL;")
        select = compiler.transform_journal(journal)
        self.assertEqual(select,
            ast.Select([
                ast.Target(ast.Column('date'), None),
                ast.Target(ast.Column('flag'), None),
                ast.Target(ast.Function('maxwidth', [
                    ast.Column('payee'), ast.Constant(48)]), None),
                ast.Target(ast.Function('maxwidth', [
                    ast.Column('narration'), ast.Constant(80)]), None),
                ast.Target(ast.Column('account'), None),
                ast.Target(ast.Column('position'), None),
                ast.Target(ast.Column('balance'), None),
            ],
            None, None, None, None, None, None, None))

    def test_journal_with_account(self):
        journal = parser.parse("JOURNAL 'liabilities';")
        select = compiler.transform_journal(journal)
        self.assertEqual(select, ast.Select([
            ast.Target(ast.Column('date'), None),
            ast.Target(ast.Column('flag'), None),
            ast.Target(ast.Function('maxwidth', [
                ast.Column('payee'),
                ast.Constant(48)]), None),
            ast.Target(ast.Function('maxwidth', [
                ast.Column('narration'),
                ast.Constant(80)]), None),
            ast.Target(ast.Column('account'), None),
            ast.Target(ast.Column('position'), None),
            ast.Target(ast.Column('balance'), None),
        ],
        None,
        ast.Match(ast.Column('account'), ast.Constant('liabilities')),
        None, None, None, None, None))

    def test_journal_with_account_and_from(self):
        journal = parser.parse("JOURNAL 'liabilities' FROM year = 2014;")
        select = compiler.transform_journal(journal)
        self.assertEqual(select, ast.Select([
            ast.Target(ast.Column('date'), None),
            ast.Target(ast.Column('flag'), None),
            ast.Target(ast.Function('maxwidth', [
                ast.Column('payee'),
                ast.Constant(48)]), None),
            ast.Target(ast.Function('maxwidth', [
                ast.Column('narration'),
                ast.Constant(80)]), None),
            ast.Target(ast.Column('account'), None),
            ast.Target(ast.Column('position'), None),
            ast.Target(ast.Column('balance'), None),
        ],
        ast.From(ast.Equal(ast.Column('year'), ast.Constant(2014)), None, None, None),
        ast.Match(ast.Column('account'), ast.Constant('liabilities')),
        None, None, None, None, None))

    def test_journal_with_account_func_and_from(self):
        journal = parser.parse("JOURNAL 'liabilities' AT cost FROM year = 2014;")
        select = compiler.transform_journal(journal)
        self.assertEqual(select, ast.Select([
            ast.Target(ast.Column('date'), None),
            ast.Target(ast.Column('flag'), None),
            ast.Target(ast.Function('maxwidth', [
                ast.Column('payee'),
                ast.Constant(48)]), None),
            ast.Target(ast.Function('maxwidth', [
                ast.Column('narration'),
                ast.Constant(80)]), None),
            ast.Target(ast.Column('account'), None),
            ast.Target(ast.Function('cost', [ast.Column('position')]), None),
            ast.Target(ast.Function('cost', [ast.Column('balance')]), None),
        ],
        ast.From(ast.Equal(ast.Column('year'), ast.Constant(2014)), None, None, None),
        ast.Match(ast.Column('account'), ast.Constant('liabilities')),
        None, None, None, None, None))


class TestTranslationBalance(CompileSelectBase):

    group_by = ast.GroupBy([
        ast.Column('account'),
        ast.Function('account_sortkey', [
            ast.Column(name='account')])], None)

    order_by = [ast.OrderBy(ast.Function('account_sortkey', [ast.Column('account')]), ast.Ordering.ASC)]

    def test_balance(self):
        balance = parser.parse("BALANCES;")
        select = compiler.transform_balances(balance)
        self.assertEqual(select, ast.Select([
            ast.Target(ast.Column('account'), None),
            ast.Target(ast.Function('sum', [
                ast.Column('position')
            ]), None),
        ],
        None, None, self.group_by, self.order_by, None, None, None))

    def test_balance_with_units(self):
        balance = parser.parse("BALANCES AT cost;")
        select = compiler.transform_balances(balance)
        self.assertEqual(select, ast.Select([
            ast.Target(ast.Column('account'), None),
            ast.Target(ast.Function('sum', [
                ast.Function('cost', [
                    ast.Column('position')
                ])
            ]), None)
        ],
        None, None, self.group_by, self.order_by, None, None, None))

    def test_balance_with_units_and_from(self):
        balance = parser.parse("BALANCES AT cost FROM year = 2014;")
        select = compiler.transform_balances(balance)
        self.assertEqual(select, ast.Select([
            ast.Target(ast.Column('account'), None),
            ast.Target(ast.Function('sum', [
                ast.Function('cost', [
                    ast.Column('position')
                ])
            ]), None),
        ],
        ast.From(ast.Equal(ast.Column('year'), ast.Constant(2014)), None, None, None),
        None, self.group_by, self.order_by, None, None, None))

    def test_print(self):
        self.assertCompile(
            qc.EvalQuery(
                Table('entries'),
                [qc.EvalTarget(qc.EvalRow(), 'ROW(*)', False)],
                None, None, None, None, None, False),
            "PRINT;",
        )

    def test_print_from(self):
        self.assertCompile(
            qc.EvalQuery(
                Table('entries'),
                [qc.EvalTarget(qc.EvalRow(), 'ROW(*)', False)],
                qc.Operator(ast.Equal, [
                    Column('year', int),
                    qc.EvalConstant(2014) ]),
                None, None, None, None, False),
            "PRINT FROM year = 2014;")


class TestCompileParameters(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.context = Connection()

    def compile(self, query, params):
        c = compiler.Compiler(self.context)
        c.table = self.context.tables.get('')
        return c.compile(parser.parse(query), params)

    def test_named_parameters(self):
        query = self.compile('''SELECT %(x)s + %(y)s''', {'x': 1, 'y': 2})
        self.assertEqual(query, qc.EvalQuery(
            Table(''), [
                # addition of constants is optimized away
                qc.EvalTarget(qc.EvalConstant(3), '%(x)s + %(y)s', False)
            ], None, None, None, None, None, None))

    def test_positional_parameters(self):
        query = self.compile('''SELECT %s + %s''', (1, 2, ))
        self.assertEqual(query, qc.EvalQuery(
            Table(''), [
                # addition of constants is optimized away
                qc.EvalTarget(qc.EvalConstant(3), '%s + %s', False)
            ], None, None, None, None, None, None))

    def test_mixing_parameters(self):
        with self.assertRaises(ProgrammingError):
            self.compile('''SELECT %s + %(foo)s''', (1, 2))

    def test_missing_parameters_positional(self):
        with self.assertRaises(ProgrammingError):
            self.compile('''SELECT %s + %s''', (1, ))

    def test_missing_parameters_named(self):
        with self.assertRaises(ProgrammingError):
            self.compile('''SELECT %(x)s + %(y)s''', {'x': 1})


class TestSelectFrom(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.conn = beanquery.connect('')
        # create a table ``postings`` with coulums ``date`` and ``account``
        cls.conn.execute('''CREATE TABLE postings (date date, account str)''')
        # make it the default table
        cls.conn.tables[None] = cls.conn.tables['postings']
        # create a table ``foo`` with columns ``x`` and ``y``
        cls.conn.execute('''CREATE TABLE foo (x int, y int)''')

    def compile(self, query):
        c = compiler.Compiler(self.conn)
        return c.compile(parser.parse(query))

    def test_select_from(self):
        query = self.compile('''SELECT x FROM foo''')
        self.assertEqual(query.table, self.conn.tables['foo'])

    def test_select_from_invalid(self):
        with self.assertRaisesRegex(beanquery.ProgrammingError, 'column "qux" not found in table "postings"'):
            self.compile('''SELECT x FROM qux''')

    def test_select_from_column(self):
        query = self.compile('''SELECT account FROM date''')
        self.assertEqual(query.table, self.conn.tables['postings'])

    def test_select_from_hash(self):
        query = self.compile('''SELECT x FROM #foo''')
        self.assertEqual(query.table, self.conn.tables['foo'])

    def test_select_from_hash_invalid(self):
        with self.assertRaisesRegex(beanquery.ProgrammingError, 'table "qux" does not exist'):
            self.compile('''SELECT x FROM #qux''')

    def test_select_from_hash_column_invalid(self):
        with self.assertRaisesRegex(beanquery.ProgrammingError, 'table "date" does not exist'):
            self.compile('''SELECT account FROM #date''')

    def test_select_from_hash_column(self):
        self.conn.execute('''CREATE table date (year int, month int, day int)''')
        def cleanup():
            del self.conn.tables['date']
        self.addCleanup(cleanup)
        query = self.compile('''SELECT year FROM #date''')
        self.assertEqual(query.table, self.conn.tables['date'])


class TestQuotedIdentifiers(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.conn = beanquery.connect('')
        # create a table ``postings`` with coulums ``date`` and ``account``
        cls.conn.execute('''CREATE TABLE postings (date date, account str)''')

    def compile(self, query):
        c = compiler.Compiler(self.conn)
        return c.compile(parser.parse(query))

    def test_from_quoted(self):
        query = self.compile('''SELECT * FROM postings''')
        self.assertIs(query.table, self.conn.tables['postings'])
        query = self.compile('''SELECT * FROM "postings"''')
        self.assertIs(query.table, self.conn.tables['postings'])

    def test_quoted_target(self):
        query = self.compile('''SELECT date FROM postings''')
        self.assertEqual(query.c_targets[0].c_expr, self.conn.tables['postings'].columns['date'])
        query = self.compile('''SELECT "date" FROM postings''')
        self.assertEqual(query.c_targets[0].c_expr, self.conn.tables['postings'].columns['date'])

    def test_quoted_string_target(self):
        # if the double quoted string is not a table name, it is a string literal ideed
        query = self.compile('''SELECT "baz" FROM postings''')
        self.assertEqual(query.c_targets[0].c_expr.value, 'baz')

    def test_quoted_in_expression(self):
        query = self.compile('''SELECT date + 1 FROM postings''')
        self.assertEqual(query.c_targets[0].c_expr.left, self.conn.tables['postings'].columns['date'])
        query = self.compile('''SELECT "date" + 1 FROM postings''')
        self.assertEqual(query.c_targets[0].c_expr.left, self.conn.tables['postings'].columns['date'])

    def test_quoted_string_in_expression(self):
        # if the double quoted string is not a table name, it is a string literal ideed
        query = self.compile('''SELECT "a" + "b" FROM postings''')
        self.assertEqual(query.c_targets[0].c_expr.value, 'ab')



__copyright__ = "Copyright (C) 2014-2016  Martin Blais"
__license__ = "GNU GPLv2"

import datetime
import textwrap
import unittest

from decimal import Decimal as D
from beanquery import parser
from beanquery.parser import ast


def Select(targets, from_clause=None, where_clause=None, **kwargs):
    defaults = dict(targets=targets,
                    from_clause=from_clause,
                    where_clause=where_clause,
                    group_by=None,
                    order_by=None,
                    pivot_by=None,
                    limit=None,
                    distinct=None)
    defaults.update(kwargs)
    return ast.Select(**defaults)


class QueryParserTestBase(unittest.TestCase):

    def parse(self, query):
        return parser.parse(query.strip())

    def assertParse(self, query, expected):
        self.assertEqual(parser.parse(query), expected)

    def assertParseTarget(self, query, expected):
        expr = parser.parse(query)
        self.assertIsInstance(expr, ast.Select)
        self.assertEqual(len(expr.targets), 1)
        self.assertEqual(expr.targets[0].expression, expected)

    def assertParseFrom(self, query, expected):
        expr = parser.parse(query)
        self.assertIsInstance(expr, ast.Select)
        self.assertEqual(expr.from_clause, expected)


class TestParseSelect(QueryParserTestBase):

    def test_select(self):
        with self.assertRaises(parser.ParseError):
            parser.parse("SELECT")

        with self.assertRaises(parser.ParseError):
            parser.parse("SELECT ; ")

        self.assertParse(
            "SELECT *;",
            Select(ast.Asterisk()))

        self.assertParse(
            "SELECT date;",
            Select([
                ast.Target(ast.Column('date'), None)
            ]))

        self.assertParse(
            "SELECT date, account",
            Select([
                ast.Target(ast.Column('date'), None),
                ast.Target(ast.Column('account'), None)
            ]))

        self.assertParse(
            "SELECT date as xdate;",
            Select([
                ast.Target(ast.Column('date'), 'xdate')
            ]))

        self.assertParse(
            "SELECT date as x, account, position as y;",
            Select([
                ast.Target(ast.Column('date'), 'x'),
                ast.Target(ast.Column('account'), None),
                ast.Target(ast.Column('position'), 'y')
            ]))

    def test_literals(self):
        # not a list
        self.assertParseTarget("SELECT (1);", ast.Constant(1))

        # list
        self.assertParseTarget("SELECT (1, );", ast.Constant([1]))
        self.assertParseTarget("SELECT (1, 2);", ast.Constant([1, 2]))
        self.assertParseTarget("SELECT (1, 2, );", ast.Constant([1, 2]))
        self.assertParseTarget("SELECT ('x', 'y', 'z');", ast.Constant(['x', 'y', 'z']))

        # column
        self.assertParseTarget("SELECT date;", ast.Column('date'))

    def test_expressions(self):
        # comparison operators
        self.assertParseTarget("SELECT a = 42;", ast.Equal(ast.Column('a'), ast.Constant(42)))
        self.assertParseTarget("SELECT a != 42;", ast.NotEqual(ast.Column('a'), ast.Constant(42)))
        self.assertParseTarget("SELECT a > 42;", ast.Greater(ast.Column('a'), ast.Constant(42)))
        self.assertParseTarget("SELECT a >= 42;", ast.GreaterEq(ast.Column('a'), ast.Constant(42)))
        self.assertParseTarget("SELECT a < 42;", ast.Less(ast.Column('a'), ast.Constant(42)))
        self.assertParseTarget("SELECT a <= 42;", ast.LessEq(ast.Column('a'), ast.Constant(42)))
        self.assertParseTarget("SELECT a ~ 'abc';", ast.Match(ast.Column('a'), ast.Constant('abc')))
        self.assertParseTarget("SELECT not a;", ast.Not(ast.Column('a')))
        self.assertParseTarget("SELECT a IS NULL;", ast.IsNull(ast.Column('a')))
        self.assertParseTarget("SELECT a IS NOT NULL;", ast.IsNotNull(ast.Column('a')))

        # bool expressions
        self.assertParseTarget("SELECT a AND b;", ast.And([ast.Column('a'), ast.Column('b')]))
        self.assertParseTarget("SELECT a AND b AND c;", ast.And([ast.Column('a'), ast.Column('b'), ast.Column('c')]))
        self.assertParseTarget("SELECT a OR b;", ast.Or([ast.Column('a'), ast.Column('b')]))
        self.assertParseTarget("SELECT a OR b OR c;", ast.Or([ast.Column('a'), ast.Column('b'), ast.Column('c')]))
        self.assertParseTarget("SELECT a AND b OR c;", ast.Or([ast.And([ast.Column('a'), ast.Column('b')]), ast.Column('c')]))
        self.assertParseTarget("SELECT NOT a;", ast.Not(ast.Column('a')))

        # math expressions with identifiers
        self.assertParseTarget("SELECT a * b;", ast.Mul(ast.Column('a'), ast.Column('b')))
        self.assertParseTarget("SELECT a / b;", ast.Div(ast.Column('a'), ast.Column('b')))
        self.assertParseTarget("SELECT a + b;", ast.Add(ast.Column('a'), ast.Column('b')))
        self.assertParseTarget("SELECT a+b;", ast.Add(ast.Column('a'), ast.Column('b')))
        self.assertParseTarget("SELECT a - b;", ast.Sub(ast.Column('a'), ast.Column('b')))
        self.assertParseTarget("SELECT a-b;", ast.Sub(ast.Column('a'), ast.Column('b')))
        self.assertParseTarget("SELECT +a;", ast.Column('a'))
        self.assertParseTarget("SELECT -a;", ast.Neg(ast.Column('a')))

        # math expressions with numerals
        self.assertParseTarget("SELECT 2 * 3;", ast.Mul(ast.Constant(2), ast.Constant(3)))
        self.assertParseTarget("SELECT 2 / 3;", ast.Div(ast.Constant(2), ast.Constant(3)))
        self.assertParseTarget("SELECT 2+(3);", ast.Add(ast.Constant(2), ast.Constant(3)))
        self.assertParseTarget("SELECT (2)-3;", ast.Sub(ast.Constant(2), ast.Constant(3)))
        self.assertParseTarget("SELECT 2 + 3;", ast.Add(ast.Constant(2), ast.Constant(3)))
        self.assertParseTarget("SELECT 2+3;", ast.Add(ast.Constant(2), ast.Constant(3)))
        self.assertParseTarget("SELECT 2 - 3;", ast.Sub(ast.Constant(2), ast.Constant(3)))
        self.assertParseTarget("SELECT 2-3;", ast.Sub(ast.Constant(2), ast.Constant(3)))
        self.assertParseTarget("SELECT +2;", ast.Constant(2))
        self.assertParseTarget("SELECT -2;", ast.Neg(ast.Constant(2)))
        # silly, fails at compile time
        self.assertParseTarget("SELECT -'abc';", ast.Neg(ast.Constant('abc')))

        # functions
        self.assertParseTarget("SELECT random();", ast.Function('random', []))
        self.assertParseTarget("SELECT min(a);", ast.Function('min', [ast.Column('a')]))
        self.assertParseTarget("SELECT min(a, b);", ast.Function('min', [ast.Column('a'), ast.Column('b')]))
        self.assertParseTarget("SELECT count(*);", ast.Function('count', [ast.Asterisk()]))

    def test_non_associative(self):
        # non associative operators
        self.assertRaises(parser.ParseError, self.parse, "SELECT 3 > 2 > 1")
        self.assertRaises(parser.ParseError, self.parse, "SELECT 3 = 2 = 1")

    def test_complex_expressions(self):
        self.assertParseTarget(
            "SELECT NOT a = (b != (42 AND 17));",
            ast.Not(
                ast.Equal(
                    ast.Column('a'),
                    ast.NotEqual(
                        ast.Column('b'),
                        ast.And([
                            ast.Constant(42),
                            ast.Constant(17)])))))


class TestSelectPrecedence(QueryParserTestBase):

    def test_operators_precedence(self):

        self.assertParseTarget(
            "SELECT a AND b OR c AND d;",
            ast.Or([ast.And([ast.Column('a'), ast.Column('b')]),
                   ast.And([ast.Column('c'), ast.Column('d')])]))

        self.assertParseTarget(
            "SELECT a = 2 AND b != 3;",
            ast.And([ast.Equal(ast.Column('a'), ast.Constant(2)),
                    ast.NotEqual(ast.Column('b'), ast.Constant(3))]))

        self.assertParseTarget(
            "SELECT not a AND b;",
            ast.And([ast.Not(ast.Column('a')), ast.Column('b')]))

        self.assertParseTarget(
            "SELECT a + b AND c - d;",
            ast.And([ast.Add(ast.Column('a'), ast.Column('b')),
                    ast.Sub(ast.Column('c'), ast.Column('d'))]))

        self.assertParseTarget(
            "SELECT a * b + c / d - 3;",
            ast.Sub(
                ast.Add(
                    ast.Mul(
                        ast.Column(name='a'),
                        ast.Column(name='b')),
                    ast.Div(ast.Column(name='c'),
                           ast.Column(name='d'))),
                ast.Constant(value=3)))

        self.assertParseTarget(
            "SELECT 'orange' IN tags AND 'bananas' IN tags;",
            ast.And([
                ast.In(ast.Constant('orange'), ast.Column('tags')),
                ast.In(ast.Constant('bananas'), ast.Column('tags'))]))


class TestSelectFrom(QueryParserTestBase):

    def test_select_from(self):
        expr = ast.Equal(ast.Column('d'), ast.And([ast.Function('max', [ast.Column('e')]), ast.Constant(17)]))

        with self.assertRaises(parser.ParseError):
            parser.parse("SELECT a, b FROM;")

        # simple
        self.assertParseFrom(
            "SELECT a, b FROM d = (max(e) and 17);",
            ast.From(expr, None, None, None))

        # open dated
        self.assertParseFrom(
            "SELECT a, b FROM d = (max(e) and 17) OPEN ON 2014-01-01;",
            ast.From(expr, datetime.date(2014, 1, 1), None, None))

        # close default
        self.assertParseFrom(
            "SELECT a, b FROM d = (max(e) and 17) CLOSE;",
            ast.From(expr, None, True, None))

        # close dated
        self.assertParseFrom(
            "SELECT a, b FROM d = (max(e) and 17) CLOSE ON 2014-10-18;",
            ast.From(expr, None, datetime.date(2014, 10, 18), None))

        # close no expression
        self.assertParseFrom(
            "SELECT a, b FROM CLOSE;",
            ast.From(None, None, True, None))

        # close no expression dated
        self.assertParseFrom(
            "SELECT a, b FROM CLOSE ON 2014-10-18;",
            ast.From(None, None, datetime.date(2014, 10, 18), None))

        # clear default
        self.assertParseFrom(
            "SELECT a, b FROM d = (max(e) and 17) CLEAR;",
            ast.From(expr, None, None, True))

        # open close clear
        self.assertParseFrom(
            "SELECT a, b FROM d = (max(e) and 17) OPEN ON 2013-10-25 CLOSE ON 2014-10-25 CLEAR;",
            ast.From(expr, datetime.date(2013, 10, 25), datetime.date(2014, 10, 25), True))


class TestSelectWhere(QueryParserTestBase):

    def test_where(self):
        expr = ast.Equal(ast.Column('d'), ast.And([ast.Function('max', [ast.Column('e')]), ast.Constant(17)]))
        self.assertParse(
            "SELECT a, b WHERE d = (max(e) and 17);",
            Select([
                ast.Target(ast.Column('a'), None),
                ast.Target(ast.Column('b'), None)
            ], None, expr))

        with self.assertRaises(parser.ParseError):
            parser.parse("SELECT a, b WHERE;")


class TestSelectFromAndWhere(QueryParserTestBase):

    def test_from_and_where(self):
        expr = ast.Equal(ast.Column('d'), ast.And([ast.Function('max', [ast.Column('e')]), ast.Constant(17)]))
        self.assertParse(
            "SELECT a, b FROM d = (max(e) and 17) WHERE d = (max(e) and 17);",
            Select([
                ast.Target(ast.Column('a'), None),
                ast.Target(ast.Column('b'), None)
            ], ast.From(expr, None, None, None), expr))


class TestSelectFromSelect(QueryParserTestBase):

    def test_from_select(self):
        self.assertParse("""
            SELECT a, b FROM (
              SELECT * FROM date = 2014-05-02
            ) WHERE c = 5 LIMIT 100;""",
            Select([
                ast.Target(ast.Column('a'), None),
                ast.Target(ast.Column('b'), None)],
            Select(
                ast.Asterisk(),
                ast.From(
                    ast.Equal(
                        ast.Column('date'),
                        ast.Constant(datetime.date(2014, 5, 2))),
                    None, None, None)),
            ast.Equal(ast.Column('c'), ast.Constant(5)),
            limit=100))


class TestSelectGroupBy(QueryParserTestBase):

    def test_groupby_one(self):
        self.assertParse(
            "SELECT * GROUP BY a;",
            Select(ast.Asterisk(),
                   group_by=ast.GroupBy([ast.Column('a')], None)))

    def test_groupby_many(self):
        self.assertParse(
            "SELECT * GROUP BY a, b, c;",
            Select(ast.Asterisk(),
                   group_by=ast.GroupBy([
                       ast.Column('a'),
                       ast.Column('b'),
                       ast.Column('c')], None)))

    def test_groupby_expr(self):
        self.assertParse(
            "SELECT * GROUP BY length(a) > 0, b;",
            Select(ast.Asterisk(),
                   group_by=ast.GroupBy([
                       ast.Greater(
                           ast.Function('length', [
                               ast.Column('a')]),
                           ast.Constant(0)),
                       ast.Column('b')], None)))

    def test_groupby_having(self):
        self.assertParse(
            "SELECT * GROUP BY a HAVING sum(x) = 0;",
            Select(ast.Asterisk(),
                   group_by=ast.GroupBy([ast.Column('a')],
                        ast.Equal(
                            ast.Function('sum', [
                                ast.Column('x')]),
                            ast.Constant(0)))))

    def test_groupby_numbers(self):
        self.assertParse(
            "SELECT * GROUP BY 1;",
            Select(ast.Asterisk(),
                   group_by=ast.GroupBy([1], None)))

        self.assertParse(
            "SELECT * GROUP BY 2, 4, 5;",
            Select(ast.Asterisk(),
                   group_by=ast.GroupBy([2, 4, 5], None)))

    def test_groupby_empty(self):
        with self.assertRaises(parser.ParseError):
            parser.parse("SELECT * GROUP BY;")


class TestSelectOrderBy(QueryParserTestBase):

    def test_orderby_one(self):
        self.assertParse(
            "SELECT * ORDER BY a;",
            Select(ast.Asterisk(),
                   order_by=[
                       ast.OrderBy(ast.Column('a'), ast.Ordering.ASC)]))

    def test_orderby_many(self):
        self.assertParse(
            "SELECT * ORDER BY a, b, c;",
            Select(ast.Asterisk(),
                   order_by=[
                       ast.OrderBy(ast.Column('a'), ast.Ordering.ASC),
                       ast.OrderBy(ast.Column('b'), ast.Ordering.ASC),
                       ast.OrderBy(ast.Column('c'), ast.Ordering.ASC)]))

    def test_orderby_asc(self):
        self.assertParse(
            "SELECT * ORDER BY a ASC;",
            Select(ast.Asterisk(),
                   order_by=[
                       ast.OrderBy(ast.Column('a'), ast.Ordering.ASC)]))

    def test_orderby_desc(self):
        self.assertParse(
            "SELECT * ORDER BY a DESC;",
            Select(ast.Asterisk(),
                   order_by=[
                       ast.OrderBy(ast.Column('a'), ast.Ordering.DESC)]))

    def test_orderby_many_asc_desc(self):
        self.assertParse(
            "SELECT * ORDER BY a ASC, b DESC, c;",
            Select(ast.Asterisk(),
                   order_by=[
                       ast.OrderBy(ast.Column('a'), ast.Ordering.ASC),
                       ast.OrderBy(ast.Column('b'), ast.Ordering.DESC),
                       ast.OrderBy(ast.Column('c'), ast.Ordering.ASC)]))

    def test_orderby_empty(self):
        with self.assertRaises(parser.ParseError):
            parser.parse("SELECT * ORDER BY;")


class TestSelectPivotBy(QueryParserTestBase):

    def test_pivotby(self):
        with self.assertRaises(parser.ParseError):
            parser.parse("SELECT * PIVOT BY;")

        with self.assertRaises(parser.ParseError):
            parser.parse("SELECT * PIVOT BY a;")

        with self.assertRaises(parser.ParseError):
            parser.parse("SELECT * PIVOT BY a, b, c")

        self.assertParse(
            "SELECT * PIVOT BY a, b",
            Select(ast.Asterisk(), pivot_by=ast.PivotBy([ast.Column('a'), ast.Column('b')])))

        self.assertParse(
            "SELECT * PIVOT BY 1, 2",
            Select(ast.Asterisk(), pivot_by=ast.PivotBy([1, 2])))


class TestSelectOptions(QueryParserTestBase):

    def test_distinct(self):
        self.assertParse(
            "SELECT DISTINCT x;", Select([ast.Target(ast.Column('x'), None)], distinct=True))

    def test_limit_present(self):
        self.assertParse(
            "SELECT * LIMIT 45;", Select(ast.Asterisk(), limit=45))

    def test_limit_empty(self):
        with self.assertRaises(parser.ParseError):
            parser.parse("SELECT * LIMIT;")


class TestBalances(QueryParserTestBase):

    def test_balances_empty(self):
        self.assertParse(
            "BALANCES;", ast.Balances(None, None, None))

    def test_balances_from(self):
        self.assertParse(
            "BALANCES FROM date = 2014-01-01 CLOSE;",
            ast.Balances(
                None,
                ast.From(
                    ast.Equal(
                        ast.Column('date'),
                        ast.Constant(datetime.date(2014, 1, 1))),
                    None, True, None),
                None))

    def test_balances_from_with_transformer(self):
        self.assertParse(
            "BALANCES AT units FROM date = 2014-01-01 CLOSE;",
            ast.Balances('units',
                ast.From(
                    ast.Equal(
                        ast.Column('date'),
                        ast.Constant(datetime.date(2014, 1, 1))),
                    None, True, None),
                None))

    def test_balances_from_with_transformer_simple(self):
        self.assertParse(
            "BALANCES AT units WHERE date = 2014-01-01;",
            ast.Balances('units',
                None,
                ast.Equal(
                    ast.Column('date'),
                    ast.Constant(datetime.date(2014, 1, 1)))))


class TestJournal(QueryParserTestBase):

    def test_journal_empty(self):
        self.assertParse(
            "JOURNAL;",
            ast.Journal(None, None, None))

    def test_journal_account(self):
        self.assertParse(
            "JOURNAL 'Assets:Checking';",
            ast.Journal('Assets:Checking', None, None))

    def test_journal_summary(self):
        self.assertParse(
            "JOURNAL AT cost;",
            ast.Journal(None, 'cost', None))

    def test_journal_account_and_summary(self):
        self.assertParse(
            "JOURNAL 'Assets:Foo' AT cost;",
            ast.Journal('Assets:Foo', 'cost', None))

    def test_journal_from(self):
        self.assertParse(
            "JOURNAL FROM date = 2014-01-01 CLOSE;",
            ast.Journal(None, None,
                ast.From(
                    ast.Equal(
                        ast.Column('date'),
                        ast.Constant(datetime.date(2014, 1, 1))
                    ), None, True, None)))


class TestPrint(QueryParserTestBase):

    def test_print_empty(self):
        self.assertParse(
            "PRINT;", ast.Print(None))

    def test_print_from(self):
        self.assertParse(
            "PRINT FROM date = 2014-01-01 CLOSE;",
            ast.Print(
                ast.From(
                    ast.Equal(
                        ast.Column('date'),
                        ast.Constant(datetime.date(2014, 1, 1))
                    ), None, True, None)))


class TestComments(QueryParserTestBase):

    def test_comments(self):
        self.assertParse(
            """SELECT first, /* comment */ second""",
            Select([
                ast.Target(ast.Column('first'), None),
                ast.Target(ast.Column('second'), None)
            ]))

        self.assertParse(
            """SELECT first, /*
                   comment
                   */ second;""",
            Select([
                ast.Target(ast.Column('first'), None),
                ast.Target(ast.Column('second'), None),
            ]))

        self.assertParse(
            """SELECT first, /**/ second;""",
            Select([
                ast.Target(ast.Column('first'), None),
                ast.Target(ast.Column('second'), None),
            ]))

        self.assertParse(
            """SELECT first, /* /* */ second;""",
            Select([
                ast.Target(ast.Column('first'), None),
                ast.Target(ast.Column('second'), None),
            ]))

        self.assertParse(
            """SELECT first, /* ; */ second;""",
            Select([
                ast.Target(ast.Column('first'), None),
                ast.Target(ast.Column('second'), None),
            ]))


class TestRepr(unittest.TestCase):

    def test_ordering(self):
        self.assertEqual(repr(ast.Ordering.ASC), 'Ordering.ASC')
        self.assertEqual(repr(ast.Ordering.DESC), 'Ordering.DESC')

    def test_ast_node(self):
        self.assertEqual(repr(ast.Constant(1)), 'Constant(value=1)')
        self.assertEqual(repr(ast.Not(ast.Constant(False))), 'Not(operand=Constant(value=False))')

    def test_tosexp(self):
        sexp = parser.parse('SELECT a + 1 FROM #test WHERE a > 42 ORDER BY b DESC').tosexp()
        self.assertEqual(sexp, textwrap.dedent('''\
            (select
              targets: (
                (target
                  expression: (add
                    left: (column
                      name: 'a')
                    right: (constant
                      value: 1))))
              from-clause: (table
                name: 'test')
              where-clause: (greater
                left: (column
                  name: 'a')
                right: (constant
                  value: 42))
              order-by: (
                (orderby
                  column: (column
                    name: 'b')
                  ordering: desc)))'''))

    def test_walk(self):
        query = parser.parse('SELECT a + 1 FROM #test WHERE a > %(foo)s ORDER BY b DESC')
        placeholders = [node for node in query.walk() if isinstance(node, ast.Placeholder)]
        self.assertEqual(placeholders, [ast.Placeholder(name='foo')])


class TestNodeText(unittest.TestCase):

    def test_text(self):
        select = parser.parse('SELECT date + 1')
        self.assertEqual(select.text, 'SELECT date + 1')
        self.assertEqual(select.targets[0].expression.text, 'date + 1')
        self.assertEqual(select.targets[0].expression.left.text, 'date')
        self.assertEqual(select.targets[0].expression.right.text, '1')

    def test_synthetic(self):
        node = ast.And([ast.Constant(False), ast.Constant(True)])
        self.assertIsNone(node.text)


class TestIdentifier(unittest.TestCase):

    @staticmethod
    def parse(string):
        return parser.BQLParser().parse(string, start='identifier', semantics=parser.BQLSemantics())

    def test_unquoted_identifer(self):
        self.assertEqual(self.parse('foo'), 'foo')
        # normalization to lower case
        self.assertEqual(self.parse('Foo'), 'foo')

    def test_quoted_identifier(self):
        self.assertEqual(self.parse('"foo"'), 'foo')
        self.assertEqual(self.parse('"foo bar"'), 'foo bar')
        self.assertEqual(self.parse('"1 + 2"'), '1 + 2')
        # no normalization to lower case
        self.assertEqual(self.parse('"Foo"'), 'Foo')
        # keywords allowed
        self.assertEqual(self.parse('"select"'), 'select')
        # quoted quotes
        self.assertEqual(self.parse('"foo""bar"'), 'foo"bar')


class TestLiteral(unittest.TestCase):

    @staticmethod
    def parse(string):
        return parser.BQLParser().parse(string, start='literal', semantics=parser.BQLSemantics())

    def assertEqualEx(self, this, that):
        self.assertIsInstance(this, type(that))
        self.assertEqual(this, that)

    def test_null(self):
        self.assertEqualEx(self.parse('NULL'), None)
        self.assertEqualEx(self.parse('null'), None)
        self.assertEqualEx(self.parse('Null'), None)

    def test_bool(self):
        self.assertEqualEx(self.parse('TRUE'), True)
        self.assertEqualEx(self.parse('true'), True)
        self.assertEqualEx(self.parse('True'), True)
        self.assertEqualEx(self.parse('FALSE'), False)
        self.assertEqualEx(self.parse('false'), False)
        self.assertEqualEx(self.parse('False'), False)

    def test_int(self):
        self.assertEqualEx(self.parse('17'), 17)

    def test_decimal(self):
        self.assertEqualEx(self.parse('17.345'), D('17.345'))
        self.assertEqualEx(self.parse('.345'), D('.345'))
        self.assertEqualEx(self.parse('17.'), D('17.'))

    def test_string(self):
        # single quoted string
        self.assertEqualEx(self.parse("'rainy-day'"), 'rainy-day')
        # single quites can be entered doubling them
        self.assertEqualEx(self.parse("'rainy''day'"), "rainy''day")
        # double quote string, kept for backward compatibility
        self.assertEqualEx(self.parse('"rainy-day"'), 'rainy-day')
        # double quotes cannot be escaped
        self.assertEqualEx(self.parse('"rainy""day"'), 'rainy')

    def test_date(self):
        self.assertEqualEx(self.parse('1972-05-28'), datetime.date(1972, 5, 28))



"""Interpreter for the query language's AST.

This code accepts the abstract syntax tree produced by the query parser,
resolves the column and function names, compiles and interpreter and prepares a
query to be run against a list of entries.
"""

from __future__ import annotations

__copyright__ = "Copyright (C) 2014-2016  Martin Blais"
__license__ = "GNU GPLv2"

import collections
import dataclasses
import datetime
import itertools
import re
import operator

from decimal import Decimal
from typing import List

from dateutil.relativedelta import relativedelta

from beanquery import cursor
from beanquery.parser import ast
from beanquery import query_execute
from beanquery import types
from beanquery import tables


MARKER = object()

FUNCTIONS = collections.defaultdict(list)
OPERATORS = collections.defaultdict(list)


class EvalNode:
    __slots__ = ('dtype',)

    def __init__(self, dtype):
        # The output data type produce by this node. This is intended to be
        # inferred by the nodes on construction.
        assert dtype is not None, "Internal erro: Invalid dtype, must be deduced."
        self.dtype = dtype

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return all(getattr(self, attr) == getattr(other, attr) for attr in self.__slots__)

    def __str__(self):
        return "{}({})".format(type(self).__name__,
                               ', '.join(repr(getattr(self, child))
                                         for child in self.__slots__))
    __repr__ = __str__

    def childnodes(self):
        """Returns the child nodes of this node.
        Yields:
          A list of EvalNode instances.
        """
        for attr in self.__slots__:
            child = getattr(self, attr)
            if isinstance(child, EvalNode):
                yield child
            elif isinstance(child, list):
                for element in child:
                    if isinstance(element, EvalNode):
                        yield element

    def __call__(self, context):
        """Evaluate this node. This is designed to recurse on its children.
        All subclasses must override and implement this method.

        Args:
          context: The evaluation object to which the evaluation need to apply.
            This is either an entry, a Posting instance, or a particular result
            set row from a sub-select. This is the provider for the underlying
            data.
        Returns:
          The evaluated value for this sub-expression tree.
        """
        raise NotImplementedError


class EvalConstant(EvalNode):
    __slots__ = ('value',)

    def __init__(self, value, dtype=None):
        super().__init__(type(value) if dtype is None else dtype)
        self.value = value

    def __call__(self, _):
        return self.value


class EvalUnaryOp(EvalNode):
    __slots__ = ('operand', 'operator')

    def __init__(self, operator, operand, dtype):
        super().__init__(dtype)
        self.operand = operand
        self.operator = operator

    def __call__(self, context):
        operand = self.operand(context)
        return self.operator(operand)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.operator!r})'


class EvalUnaryOpSafe(EvalUnaryOp):

    def __call__(self, context):
        operand = self.operand(context)
        if operand is None:
            return None
        return self.operator(operand)


class EvalBinaryOp(EvalNode):
    __slots__ = ('left', 'right', 'operator')

    def __init__(self, operator, left, right, dtype):
        super().__init__(dtype)
        self.operator = operator
        self.left = left
        self.right = right

    def __call__(self, context):
        left = self.left(context)
        if left is None:
            return None
        right = self.right(context)
        if right is None:
            return None
        return self.operator(left, right)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.left!r}, {self.right!r})'


class EvalBetween(EvalNode):
    __slots__ = ('operand', 'lower', 'upper')

    def __init__(self, operand, lower, upper):
        super().__init__(bool)
        self.operand = operand
        self.lower = lower
        self.upper = upper

    def __call__(self, context):
        operand = self.operand(context)
        if operand is None:
            return None
        lower = self.lower(context)
        if lower is None:
            return None
        upper = self.upper(context)
        if upper is None:
            return None
        return lower <= operand <= upper


def unaryop(op, intypes, outtype, nullsafe=False):
    def decorator(func):
        class Op(EvalUnaryOp if nullsafe else EvalUnaryOpSafe):
            __intypes__ = intypes
            def __init__(self, operand):
                super().__init__(func, operand, outtype)
        Op.__name__ = f'{op.__name__}[{intypes[0].__name__}]'
        OPERATORS[op].append(Op)
        return func
    return decorator


def binaryop(op, intypes, outtype):
    def decorator(func):
        class Op(EvalBinaryOp):
            __intypes__ = intypes
            def __init__(self, left, right):
                super().__init__(func, left, right, outtype)
        Op.__name__ = f'{op.__name__}[{intypes[0].__name__},{intypes[1].__name__}]'
        OPERATORS[op].append(Op)
        return func
    return decorator


def Operator(op, operands):
    op = types.function_lookup(OPERATORS, op, operands)
    if op is not None:
        return op(*operands)
    raise KeyError


unaryop(ast.Not, [types.Any], bool, nullsafe=True)(operator.not_)

@unaryop(ast.Neg, [int], int)
@unaryop(ast.Neg, [Decimal], Decimal)
def neg_(x):
    return -x


@unaryop(ast.IsNull, [types.Any], bool, nullsafe=True)
def null(x):
    return x is None


@unaryop(ast.IsNotNull, [types.Any], bool, nullsafe=True)
def not_null(x):
    return x is not None


@binaryop(ast.Mul, [Decimal, Decimal], Decimal)
@binaryop(ast.Mul, [Decimal, int], Decimal)
@binaryop(ast.Mul, [int, Decimal], Decimal)
@binaryop(ast.Mul, [int, int], int)
def mul_(x, y):
    return x * y


@binaryop(ast.Div, [Decimal, Decimal], Decimal)
@binaryop(ast.Div, [Decimal, int], Decimal)
@binaryop(ast.Div, [int, Decimal], Decimal)
def div_(x, y):
    if y == 0:
        return None
    return x / y


@binaryop(ast.Div, [int, int], Decimal)
def div_int(x, y):
    if y == 0:
        return None
    return Decimal(x) / y


@binaryop(ast.Mod, [int, int], int)
@binaryop(ast.Mod, [Decimal, int], Decimal)
@binaryop(ast.Mod, [int, Decimal], Decimal)
@binaryop(ast.Mod, [Decimal, Decimal], Decimal)
def mod_(x, y):
    if y == 0:
        return None
    return x % y


@binaryop(ast.Add, [Decimal, Decimal], Decimal)
@binaryop(ast.Add, [Decimal, int], Decimal)
@binaryop(ast.Add, [int, Decimal], Decimal)
@binaryop(ast.Add, [int, int], int)
@binaryop(ast.Add, [str, str], str)
@binaryop(ast.Add, [datetime.date, relativedelta], datetime.date)
@binaryop(ast.Add, [relativedelta, datetime.date], datetime.date)
@binaryop(ast.Add, [relativedelta, relativedelta], relativedelta)
def add_(x, y):
    return x + y


@binaryop(ast.Sub, [Decimal, Decimal], Decimal)
@binaryop(ast.Sub, [Decimal, int], Decimal)
@binaryop(ast.Sub, [int, Decimal], Decimal)
@binaryop(ast.Sub, [int, int], int)
@binaryop(ast.Sub, [datetime.date, relativedelta], datetime.date)
@binaryop(ast.Sub, [relativedelta, datetime.date], datetime.date)
@binaryop(ast.Sub, [relativedelta, relativedelta], datetime.date)
def sub_(x, y):
    return x - y


@binaryop(ast.Add, [datetime.date, int], datetime.date)
def add_date_int(x, y):
    return x + datetime.timedelta(days=y)


@binaryop(ast.Add, [int, datetime.date], datetime.date)
def add_int_date(x, y):
    return y + datetime.timedelta(days=x)


@binaryop(ast.Sub, [datetime.date, int], datetime.date)
def sub_date_int(x, y):
    return x - datetime.timedelta(days=y)


@binaryop(ast.Sub, [datetime.date, datetime.date], int)
def sub_date_date(x, y):
    return (x - y).days


@binaryop(ast.Match, [str, str], bool)
def match_(x, y):
    return bool(re.search(y, x, re.IGNORECASE))


@binaryop(ast.NotMatch, [str, str], bool)
def not_match_(x, y):
    return not bool(re.search(y, x, re.IGNORECASE))


@binaryop(ast.Matches, [str, str], bool)
def matches_(x, y):
    return bool(re.search(x, y))


@binaryop(ast.In, [types.Any, set], bool)
@binaryop(ast.In, [types.Any, list], bool)
@binaryop(ast.In, [types.Any, dict], bool)
def in_(x, y):
    return operator.contains(y, x)


@binaryop(ast.NotIn, [types.Any, set], bool)
@binaryop(ast.NotIn, [types.Any, list], bool)
@binaryop(ast.NotIn, [types.Any, dict], bool)
def not_in_(x, y):
    return not operator.contains(y, x)


_comparisons = [
    (ast.Equal, operator.eq),
    (ast.NotEqual, operator.ne),
    (ast.Greater, operator.gt),
    (ast.GreaterEq, operator.ge),
    (ast.Less, operator.lt),
    (ast.LessEq, operator.le),
]

_intypes = [
    [int, int],
    [Decimal, int],
    [int, Decimal],
    [Decimal, Decimal],
    [datetime.date, datetime.date],
    [str, str],
]

for node, op in _comparisons:
    for intypes in _intypes:
        binaryop(node, intypes, bool)(op)

_comparable = [
    # lists of types that can be compared with each other
    [int, Decimal],
    [datetime.date],
    [str],
]

for comparable in _comparable:
    for intypes in itertools.product(comparable, repeat=3):
        class Between(EvalBetween):
            __intypes__ = list(intypes)
        OPERATORS[ast.Between].append(Between)


class EvalAnd(EvalNode):
    __slots__ = ('args',)

    def __init__(self, args):
        super().__init__(bool)
        self.args = args

    def __call__(self, context):
        for arg in self.args:
            value = arg(context)
            if value is None:
                return None
            if not value:
                return False
        return True


class EvalOr(EvalNode):
    __slots__ = ('args',)

    def __init__(self, args):
        super().__init__(bool)
        self.args = args

    def __call__(self, context):
        r = False
        for arg in self.args:
            value = arg(context)
            if value is None:
                r = None
            if value:
                return True
        return r


class EvalCoalesce(EvalNode):
    __slots__ = ('args',)

    def __init__(self, args):
        super().__init__(args[0].dtype)
        self.args = args

    def __call__(self, context):
        for arg in self.args:
            value = arg(context)
            if value is not None:
                return value
        return None


class EvalFunction(EvalNode):
    __slots__ = ('operands',)

    # Type constraints on the input arguments.
    __intypes__ = []

    def __init__(self, context, operands, dtype):
        super().__init__(dtype)
        self.context = context
        self.operands = operands


class EvalGetItem(EvalNode):
    __slots__ = ('operand', 'key')

    def __init__(self, operand, key):
        super().__init__(object)
        self.operand = operand
        self.key = key

    def __call__(self, context):
        operand = self.operand(context)
        if operand is None:
            return None
        return operand.get(self.key)


class EvalGetter(EvalNode):
    __slots__ = ('operand', 'getter')

    def __init__(self, operand, getter, dtype):
        super().__init__(dtype)
        self.operand = operand
        self.getter = getter

    def __call__(self, context):
        operand = self.operand(context)
        if operand is None:
            return None
        return self.getter(operand)


class EvalAny(EvalNode):
    __slots__ = ('op', 'left', 'right')

    def __init__(self, op, left, right):
        super().__init__(bool)
        self.op = op
        self.left = left
        self.right = right

    def __call__(self, row):
        left = self.left(row)
        if left is None:
            return None
        right = self.right(row)
        if right is None:
            return None
        return any(self.op(left, x) for x in right)


class EvalAll(EvalNode):
    __slots__ = ('op', 'left', 'right')

    def __init__(self, op, left, right):
        super().__init__(bool)
        self.op = op
        self.left = left
        self.right = right

    def __call__(self, row):
        left = self.left(row)
        if left is None:
            return None
        right = self.right(row)
        if right is None:
            return None
        return all(self.op(left, x) for x in right)


class EvalRow(EvalNode):
    __slots__ = ()

    def __init__(self):
        super().__init__(object)

    def __call__(self, context):
        return context


class EvalColumn(EvalNode):
    pass


class EvalAggregator(EvalFunction):
    pure = False

    def __init__(self, context, operands, dtype=None):
        super().__init__(context, operands, dtype or operands[0].dtype)
        self.value = None

    def allocate(self, allocator):
        """Allocate handles to store data for a node's aggregate storage.

        This is called once before beginning aggregations. If you need any
        kind of per-aggregate storage during the computation phase, get it
        in this method.

        Args:
          allocator: An instance of Allocator, on which you can call allocate() to
            obtain a handle for a slot to store data on store objects later on.
        """
        self.handle = allocator.allocate()

    def initialize(self, store):
        """Initialize this node's aggregate data.

        Args:
          store: An object indexable by handles appropriated during allocate().
        """
        store[self.handle] = self.dtype()
        self.value = None

    def update(self, store, context):
        """Evaluate this node. This is designed to recurse on its children.

        Args:
          store: An object indexable by handles appropriated during allocate().
          context: The object to which the evaluation need to apply (see __call__).
        """
        # Do nothing by default.

    def finalize(self, store):
        """Finalize this node's aggregate data.

        Args:
          store: An object indexable by handles appropriated during allocate().
        """
        self.value = store[self.handle]

    def __call__(self, context):
        """Return the value on evaluation.

        Args:
          context: The evaluation object to which the evaluation need to apply.
        Returns:
          The final aggregated value.
        """
        return self.value


class SubqueryTable(tables.Table):
    def __init__(self, subquery):
        self.columns = {}
        self.subquery = subquery
        for i, target in enumerate(target for target in subquery.c_targets if target.name is not None):
            column = self.column(i, target.name, target.c_expr.dtype)
            self.columns[target.name] = column()

    @staticmethod
    def column(i, name, dtype):
        class Column(EvalColumn):
            def __init__(self):
                super().__init__(dtype)
            __call__ = staticmethod(operator.itemgetter(i))
        return Column

    def __iter__(self):
        columns, rows = self.subquery()
        return iter(rows)


class EvalConstantSubquery1D(EvalNode):
    def __init__(self, subquery):
        self.dtype = List[subquery.columns[0].c_expr.dtype]
        self.subquery = subquery
        self.value = MARKER

    def __call__(self, context):
        if self.value is MARKER:
            # Subqueries accessing the current row are not supported yet,
            # thus the subquery result can be simply cached.
            columns, rows = self.subquery()
            value = [row[0] for row in rows]
            # Subqueries not returning any row are threates as NULL.
            self.value = value if value else None
        return self.value


# A compiled target.
#
# Attributes:
#   c_expr: A compiled expression tree (an EvalNode root node).
#   name: The name of the target. If None, this is an invisible
#     target that gets evaluated but not displayed.
#   is_aggregate: A boolean, true if 'c_expr' is an aggregate.
EvalTarget = collections.namedtuple('EvalTarget', 'c_expr name is_aggregate')


# A compiled query, ready for execution.
#
# Attributes:
#   c_targets: A list of compiled targets (instancef of EvalTarget).
#   c_where: An instance of EvalNode, a compiled expression tree, for postings.
#   group_indexes: A list of integers that describe which target indexes to
#     group by. All the targets referenced here should be non-aggregates. In fact,
#     this list of indexes should always cover all non-aggregates in 'c_targets'.
#     And this list may well include some invisible columns if only specified in
#     the GROUP BY clause.
#   order_spec: A list of (integer indexes, sort order) tuples.
#     This list may refer to either aggregates or non-aggregates.
#   limit: An optional integer used to cut off the number of result rows returned.
#   distinct: An optional boolean that requests we should uniquify the result rows.
@dataclasses.dataclass
class EvalQuery:
    table: tables.Table
    c_targets: list
    c_where: EvalNode
    group_indexes: list[int]
    having_index: int
    order_spec: list[tuple[int, ast.Ordering]]
    limit: int
    distinct: bool

    @property
    def columns(self):
        return [t for t in self.c_targets if t.name is not None]

    def __call__(self):
        return query_execute.execute_select(self)


@dataclasses.dataclass
class EvalPivot:
    """Implement PIVOT BY clause."""

    query: EvalQuery
    pivots: list[int]

    def __call__(self):
        columns, rows = self.query()

        col1, col2 = self.pivots
        othercols = [i for i in range(len(columns)) if i not in self.pivots]
        nother = len(othercols)
        other = lambda x: tuple(x[i] for i in othercols)
        keys = sorted({row[col2] for row in rows})

        # Compute the new column names and dtypes.
        if nother > 1:
            it = itertools.product(keys, other(columns))
            names = [f'{columns[col1].name}/{columns[col2].name}'] + [f'{key}/{col.name}' for key, col in it]
        else:
            names = [f'{columns[col1].name}/{columns[col2].name}'] + [f'{key}' for key in keys]
        datatypes = [columns[col1].datatype] + [col.datatype for col in other(columns)] * len(keys)
        columns = tuple(cursor.Column(name, datatype) for name, datatype in zip(names, datatypes))

        # Populate the pivoted table.
        pivoted = []
        rows.sort(key=operator.itemgetter(col1))
        for field1, group in itertools.groupby(rows, key=operator.itemgetter(col1)):
            outrow = [field1] + [None] * (len(columns) - 1)
            for row in group:
                index = keys.index(row[col2]) * nother + 1
                outrow[index:index+nother] = other(row)
            pivoted.append(tuple(outrow))

        return columns, pivoted


@dataclasses.dataclass
class EvalCreateTable:
    context: object = dataclasses.field(repr=False)
    name: str
    columns: list[tuple[str, str]]
    using: str
    data: EvalQuery | None
    create: object = dataclasses.field(repr=False)

    def __call__(self):
        table = self.create(self.name, self.columns, self.using)
        if self.data is not None:
            columns, rows = self.data()
            for row in rows:
                table.insert(row)
        self.context.tables[self.name] = table
        return (), []


@dataclasses.dataclass
class EvalInsert:
    table: tables.Table
    values: list[EvalNode]

    def __call__(self):
        values = tuple(value(None) for value in self.values)
        self.table.insert(values)
        return (), []
